/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: WindowsInterface.cpp,v 1.3 2001/03/14 13:35:48 wouter Exp $ 
 */
 
/* WindowsInterface.cpp: implementation of the WindowsInterface class.
*/

#include "WindowsInterface.h"
#include "../../lib/MobilizerPreferences.h"
#include "../../lib/PointerList.h"
#include "../../lib/String.h"
#include "../../renderer/GUIPeer.h"
#include "../../renderer/TestElementPeer.h"
#include "../../renderer/WBMP.h"
#include "../../renderer/WMLGUIElement.h"
#include "../../renderer/WMLWindow.h"
#include "TextFunctions.h"
#include "WindowsInputElementPeer.h"
#include "WindowsSelectElementPeer.h"
#include "WindowsWBMPPeer.h"

/** 
 * Constructs a new WindowsInterface object.
 * @param hwnd  the handle to the window
 * @param hFont the handle to the font
 * @param hFont the handle to the instance
 */
WindowsInterface::WindowsInterface(HWND hwnd, HINSTANCE hInstance, HFONT hFont) {

  windowHandle = hwnd;
  deviceContext = NULL;
  memoryBlock = NULL;
  currentStyle = 0;
  font = hFont;
  instance = hInstance;

}

/**
 * Delete the object
 */
WindowsInterface::~WindowsInterface() {
}


/**
 * Tells the WindowsInterface the location of the mobilizer preferences
 *
 * @param mp the pointer to the MobilizerPreferences object
 */
void WindowsInterface::setPreferences(MobilizerPreferences* mp) {

  prefs = mp;

}

/** 
 * Tells the WindowsInterface that WMLWindow is about to start drawing.
 * After this member has been called, calls to the other members can 
 * be expected. When width and height are 0, actual drawing wont take
 * place, so WindowsInterface::drawText wont be called before 
 * WindowsInterface::stopDrawing.
 *
 * @param width   the width of the content
 * @param height  the height of the content
 */
void WindowsInterface::startDrawing(int width, int height) {

  windowWidth = width;
  windowHeight = height;

  if(windowWidth != 0) {  /* drawText can be called */

    deviceContext = BeginPaint(windowHandle, &paintStruct);

  }

  memoryBlock = CreateCompatibleDC(deviceContext);
  HBITMAP memBM;

  if(windowWidth != 0) {  /* drawText can be called */

    memBM = CreateCompatibleBitmap(deviceContext,width,height);
    SelectObject(memoryBlock,memBM);
    DeleteObject(memBM);
  
    DeleteObject(SelectObject(memoryBlock,GetStockObject(WHITE_PEN)));
    Rectangle(memoryBlock,0,0,width,height);
    DeleteObject(SelectObject(memoryBlock,GetStockObject(BLACK_PEN)));

  }

  SelectObject(memoryBlock, createFont(memoryBlock,"Arial",90,0,0,TRUE));
  currentStyle = 0;

}

/** 
 * Tells the WindowDrawingInterface that WMLWindow is about to stop 
 * drawing. After this member has been called, only 
 * WindowsInterface::startDrawing can be called.
 */
void WindowsInterface::stopDrawing() {

  DeleteObject(SelectObject(memoryBlock,GetStockObject(SYSTEM_FONT)));
  if(windowWidth != 0) { /* drawText has been called */

    BitBlt(deviceContext,0,0,windowWidth,windowHeight,memoryBlock,0,0,SRCCOPY); 
    EndPaint(windowHandle, &paintStruct);

  }
  DeleteDC(memoryBlock);
  memoryBlock = NULL;
  deviceContext = NULL;

}

/**
 * Returns the height of the string in the given WMLWindow::style in
 * pixels.
 *
 * @param  s      the string
 * @param  style  the WMLWindow::style in which the string will be 
 * displayed
 * @return the height in pixels
 */
int WindowsInterface::getStringHeight(String *s, int style) {
  SIZE size;
  selectFont(style);
  SetTextJustification (memoryBlock, 0, 0) ;
  GetTextExtentPoint32 (memoryBlock, s->getPtr(), s->getLength(), &size);

  return size.cy;
}

/**
 * Returns the width of the string in the given WMLWindow::style in 
 * pixels.
 *
 * @param  s      the string
 * @param  style  the WMLWindow::style in which the string will be
 * displayed
 * @return the width in pixels
 */
int WindowsInterface::getStringWidth(String *s, int style) {
  SIZE size;
  selectFont(style);
  SetTextJustification (memoryBlock, 0, 0) ;
  GetTextExtentPoint32 (memoryBlock, s->getPtr(), s->getLength(), &size);
  
  return size.cx;

}

/** 
 * Creates a new input element. This member must return a pointer to a
 * newly allocated platform dependent object that is subclassed from 
 * the GUIPeer class. The width and height members of this class must
 * be set to the correct values, meaning: the width and height of the 
 * input element.
 * @return the platform dependent object
 */
GUIPeer* WindowsInterface::createInputElement() {
  
  return new WindowsInputElementPeer(windowHandle,instance,font);
  
}

/** 
 * Draws an input element in the display area assosiated with this 
 * object. The WMLGUIElement in the argument has members like 
 * WMLGUIElement::getHeight() and WMLGUIElement::getXCoordinate().
 *
 * @param wie the WMLGUIElement
 */
void WindowsInterface::drawInputElement(WMLGUIElement *wie) {

  MoveWindow((HWND)wie->getPeer()->getPlatformDependentPointer(),
                   wie->getXCoordinate(), 
                   wie->getYCoordinate(), 
                   wie->getWidth(),
                   wie->getHeight(), true);

}

/** 
 * Creates a new select element. This member must return a pointer
 * to a newly allocated platform dependent object that is subclassed
 * from the GUIPeer class. The width and height members of this class
 * must be set to the correct values, meaning: the width and height 
 * of the select element.
 * @return the platform dependent object
 */
GUIPeer* WindowsInterface::createSelectElement(PointerList *optionList) {

  return new WindowsSelectElementPeer(windowHandle, instance, font, optionList);

}

/** 
 * Draws a select element in the display area assosiated with this 
 * object. The WMLGUIElement in the argument has members like 
 * WMLGUIElement::getHeight() and WMLGUIElement::getXCoordinate().
 *
 * @param wse the WMLGUIElement
 */
void WindowsInterface::drawSelectElement(WMLGUIElement *wse) {

  MoveWindow((HWND)wse->getPeer()->getPlatformDependentPointer(),
                   wse->getXCoordinate(), 
                   wse->getYCoordinate(), 
                   wse->getWidth(),
                   wse->getHeight(), true);

}

/**
 * Returns the amount of pixels below the baseline of the font used
 * with the given WMLWindow::style.
 * @param  style  the WMLWindow::style from which the pixel below
 * baseline must be determined
 * @return amount of pixels below the baseline
 */
int WindowsInterface::getPixelsBelowBaseLine(int style) {

  selectFont(style);
  TEXTMETRIC tm;
  GetTextMetrics(memoryBlock, &tm);
  return tm.tmDescent;

}

/** 
 * Draws the specified text in the display area assosiated with this
 * object.
 *
 * @param text    the to be displayed text
 * @param x       the x coordinate of the left-upper point at which to
 * display text
 * @param y       the y coordinate of the left-upper point at which to 
 * display text
 * @param style   the WMLWindow::style that must be used to display text
 */
void WindowsInterface::drawText(String *text, int x, int y, int style) {

  selectFont(style);
  TextOut(memoryBlock,x,y,text->getPtr(),text->getLength());

}

/**
 * Changes the current WMLWindow::style that WindowsInterface uses. Does 
 * nothing when style == currentStyle.
 *
 * @param style the new style
 */
void WindowsInterface::selectFont(int style) {

  if(style != currentStyle) {

    HFONT hFont;
    int EZstyle = 0;
    int size = 90;

    if(style & WMLWindow::styleItalic) 
      EZstyle |= ATTR_ITALIC;

    if(style & WMLWindow::styleUnderlined || style & WMLWindow::styleLink) 
      EZstyle |= ATTR_UNDERLINE;

    if(style & WMLWindow::styleBold) 
      EZstyle |= ATTR_BOLD;

    if(style & WMLWindow::styleBig) 
      size += 40;

    if(style & WMLWindow::styleSmall) 
      size -= 25;

    hFont = createFont(memoryBlock,"Arial",size,0,EZstyle,TRUE);

    DeleteObject(SelectObject(memoryBlock,hFont));

    if(style & WMLWindow::styleLink) {

      if(style & WMLWindow::styleToplink)
        SetTextColor(memoryBlock,prefs->getPreference(MobilizerPreferences::prefTopLinkColor));
      else
        SetTextColor(memoryBlock,prefs->getPreference(MobilizerPreferences::prefLinkColor));
      
    }
    else
      SetTextColor(memoryBlock,RGB(0,0,0));

    currentStyle = style;

  }

}

GUIPeer *WindowsInterface::createImage(WBMP *image) {

  return new WindowsWBMPPeer(image);

}

/** 
 * Draws the specified image in the display area assosiated with this 
 * object. The WMLGUIElement in the argument has members like 
 * WMLGUIElement::getHeight() and WMLGUIElement::getXCoordinate().
 *
 * @param image the WMLGUIElement
 */
void WindowsInterface::drawImage(WMLGUIElement *image) {

  drawBitmap(memoryBlock, 
             (HBITMAP)image->getPeer()->getPlatformDependentPointer(),
             image->getXCoordinate(),
             image->getYCoordinate());

}

/** 
 * Draws a bitmap at (xStart, yStart) in the device context pointed to
 * by hdc.
 *
 * @param hdc     the device context to write to
 * @param hBitmap the handle to the bitmap
 * @param xStart  the x start position
 * @param ySTart  the y start position
 */
void WindowsInterface::drawBitmap(HDC hdc, HBITMAP hBitmap, int xStart, int yStart) {
	BITMAP bm;
	HDC hdcMem;
	POINT ptSize, ptOrg;

	hdcMem = CreateCompatibleDC(hdc);
	SelectObject(hdcMem, hBitmap);
	SetMapMode(hdcMem, GetMapMode(hdc));

	GetObject(hBitmap, sizeof(BITMAP), (LPVOID) &bm);

	ptSize.x = bm.bmWidth;
	ptSize.y = bm.bmHeight;
	DPtoLP (hdc, &ptSize, 1);

	ptOrg.x = 0;
	ptOrg.y = 0;
	DPtoLP(hdc, &ptOrg, 1);
	
	BitBlt(hdc, xStart, yStart, ptSize.x, ptSize.y,
    hdcMem,ptOrg.x,ptOrg.y,SRCCOPY);
	
	DeleteDC(hdcMem);

}

