/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: AbstractSocket.cpp,v 1.4 2001/03/07 11:47:39 pascal Exp $ 
 */

#include <stdlib.h>
#include <string.h>

#include "String.h"
#include "EOFException.h"
#include "OutOfMemoryException.h"
#include "AbstractSocket.h"

/**
 * Constructor; sets up internal buffers.
 */
AbstractSocket::AbstractSocket() {

  /* Create internal buffer. */
  
  bufferPos = 0;
  bufferLength = 1024;
  bufferAvailable = 0;
  buffer = (char *) malloc(bufferLength);
  stringBuffer = new String();

}

/**
 * Destructors; releases memory held by internal buffers.
 */
AbstractSocket::~AbstractSocket() {

  delete stringBuffer;
  free(buffer);

}

/**
 * Returns true if EOF has been reached on this socket.
 * @return true if EOF has been reached.
 */
bool AbstractSocket::isEOF() {

  return eof;

}

/**
 * Resets the internal buffer; that is, characters read up until
 * now are discarded and memory is freed. Needs to be called regularly,
 * or the memory will grow unbounded.
 */
void AbstractSocket::reset() {
  char *np;

  /* Save unread data in buffer. */

  memmove(buffer, buffer + bufferPos, bufferAvailable - bufferPos);

  bufferAvailable -= bufferPos;
  bufferPos = 0;

  /* Shrink buffer. */

  np = (char *) realloc(buffer, bufferAvailable > 1024 ? 
                        bufferAvailable : 1024);
  
  if(np) {

    bufferLength = bufferAvailable > 1024 ? bufferAvailable : 1024;
    buffer = np;

  }
  
}

/**
 * Reads a line (\n terminated string) from the socket and returns
 * that line as a pointer to a String object. The returned string
 * includes the final newline. Errors are thrown as exceptions.
 * @throws EOFException when end of file is reached.
 * @throws OutOfMemoryException when the internal buffer could not be
 * grown.
 * @return pointer to String object.
 */
String *AbstractSocket::readLine() {
  char c;

  stringBuffer->clear();

  try {

    c = getChar();
    stringBuffer->append(c);

    while(c != '\n') {

      c = getChar();
      stringBuffer->append(c);

    }

  } catch(EOFException& e) {
    
    reset();
    throw e;

  } catch(OutOfMemoryException& e) {

    reset();
    throw e;

  }

  return stringBuffer;

}

/**
 * Reads count bytes into buffer and blocks until that number
 * of bytes have been actually read or until EOF is reached.
 */
size_t AbstractSocket::readFully(void *buffer, size_t count) {
  unsigned int i;
  char *buf = (char *) buffer;

  try {

    for(i = 0; i < count; i++)
      buf[i] = getChar();

  } catch(Exception&) {

  }

  reset();

  return i;

}

/**
 * Writes a string.
 */
void AbstractSocket::writeString(const String *s) {

  /* Should check bytes actually written etc. */
  writeNative((void *) s->getPtr(), strlen(s->getPtr()));

}

/**
 * Reads a single character from the socket and returns it.
 * @throws EOFException when end of file is reached.
 * @throws OutOfMemoryException when the internal buffer could not be
 * grown.
 * @return character read.
 */
char AbstractSocket::getChar() {
  size_t r = 0;
  char *np;
  char c = 0;

  if(bufferPos < bufferAvailable) {

    /* Get character from buffer. */

    c = buffer[bufferPos];
    bufferPos++;

  } else {
    
    /* Buffer exhausted, try to get more bytes. */
    
    if(bufferAvailable == bufferLength) {

      /* Need to grow buffer. */

      np = (char *) realloc(buffer, bufferLength + 1024);
      
      if(np)
	bufferLength = bufferLength + 1024;
      
    } else {
      
      np = buffer;
      
    }

    if(np) {

      /* Buffer grown, now fill it. */
      
      buffer = np;

      r = readNative(buffer + bufferPos, 
		     bufferLength - bufferPos);
      bufferAvailable += r;
      
      if(r == 0) {
	
	/* EOF reached. */

	eof = true;
	throw EOFException();
	
	/* NOT REACHED */

      }

      c = buffer[bufferPos];
      bufferPos++;

    } else {
      
      /* Could not grow buffer. */
      
      throw OutOfMemoryException();

      /* NOT REACHED */
      
    }
    
  }
 
  return c;

}

