/*
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 *
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: PointerList.cpp,v 1.8 2001/03/24 07:26:51 pascal Exp $
 */

#include <malloc.h>
#include <stdlib.h>

#include "PointerList.h"

/**
 * Constructs a new PointerList.
 */
PointerList::PointerList() {

  size = 0;
  elements = NULL;

}

/**
 * Delete the object.
 */
PointerList::~PointerList() {

  delete elements;

}

/** 
 * Constructs a new PointerList object that contains the same pointers 
 * as the list in the argument.
 * @param pl the object to be copied
 */
PointerList::PointerList(PointerList* pl) {

  size = pl->getSize();
  elements = (void **)malloc(size * sizeof(void *));

  unsigned int i;
  void **otherElements = pl->getArray();

  for(i = 0; i < size; i++)
    elements[i] = otherElements[i];

}

/**
 * This function tests the code of this class.
 *
 * @return true when all tests succeed, otherwise false
 */
bool PointerList::test() {

  PointerList p;

  p.add(new int(1));
  p.add(new int(2));
  p.add(new int(3));

  PointerList l(&p);

  if(l.getSize() != 3)
    return false;

  if(*((int *)l.getFirstAdded()) != 1)
    return false;
  if(*((int *)l.getLastAdded())  != 3)
    return false;

  free(l.removeFirstAdded());

  if(l.getSize() != 2)
    return false;

  if(*((int *)l.getFirstAdded()) != 2)
    return false;
  if(*((int *)l.getLastAdded())  != 3)
    return false;

  free(l.removeLastAdded());

  if(l.getSize() != 1)
    return false;

  if(*((int *)l.getFirstAdded()) != 2)
    return false;
  if(*((int *)l.getLastAdded())  != 2)
    return false;

  free(l.removeLastAdded());

  if(l.getSize() != 0)
    return false;

  if(l.getFirstAdded() != NULL)
    return false;
  if(l.getLastAdded()  != NULL)
    return false;

  l.add(new int(4));

  if(l.getSize() != 1)
    return false;

  if(*((int *)l.getFirstAdded()) != 4)
    return false;
  if(*((int *)l.getLastAdded())  != 4)
    return false;

  free(l.removeFirstAdded());

  if(l.getSize() != 0)
    return false;

  if(l.getFirstAdded() != NULL)
    return false;
  if(l.getLastAdded()  != NULL)
    return false;

  l.add(new int(1));
  l.add(new int(2));
  l.add(new int(3));

  while(l.getSize())
    delete (int *)l.removeLastAdded();

  return true;

}

/**
 * Adds a pointer to the list
 *
 * @param   pointer   the pointer that must be added
 */
void PointerList::add(void *pointer) {

  elements = (void **)realloc(elements,(size+1) * sizeof(void *));

  elements[size] = (void *)pointer;

  size++;

}


/**
 * Returns the last added pointer from the list.
 * @return the last added pointer from the list
 */
void *PointerList::getLastAdded() {

  if(size == 0)
    return NULL;
  return elements[size - 1];

}

/**
 * Returns the first added pointer from the list.
 * @return the first added pointer from the list
 */
void *PointerList::getFirstAdded() {

  if(size == 0)
    return NULL;
  return elements[0];

}

/**
 * Removes the last added pointer from the list.
 * @return the removed pointer
 */
void *PointerList::removeLastAdded() {

  void *returnValue = NULL;

  if(size == 0)
    return NULL;

  if(size == 1) {
    returnValue = elements[0];
    delete elements;
    elements = NULL;
    size = 0;
    return returnValue;
  }

  size--;

  returnValue = elements[size];

  elements = (void **)realloc(elements,size * sizeof(void *));

  return returnValue;

}

/**
 * Removes the first added pointer from the list.
 * @return the removed pointer
 */
void *PointerList::removeFirstAdded() {

  void *returnValue = NULL;

  if(size == 0)
    return NULL;

  if(size == 1) {
    returnValue = elements[0];
    delete elements;
    elements = NULL;
    size = 0;
    return returnValue;
  }

  size--;

  returnValue = elements[0];
  void **oldElements = elements; /* remember old values */
  elements = (void **) malloc((size) * sizeof(void *)); /* allocate new memory */

  unsigned int i;
  for(i = 0; i < size; i++)
    elements[i] = oldElements[i + 1]; /* copy old values to new memory */

  delete oldElements;

  return returnValue;

}

/**
 * Returns the size of the list.
 * @return the size of the list
 */
unsigned int PointerList::getSize() {

  return size;

}

/**
 * Returns the used array.
 * @return the used array
 */
void **PointerList::getArray() {

  return elements;

}

