/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLTable.cpp,v 1.10 2001/02/28 14:10:11 wouter Exp $
 */

#include "WMLTable.h"
#include "../lib/String.h"
#include "../wmlparser/XMLAttributeList.h"
#include "../wmlparser/XMLBody.h"
#include "../wmlparser/XMLCharData.h"
#include "../wmlparser/XMLElement.h"
#include "WBMP.h"
#include "WindowDrawingInterface.h"
#include "WMLWindow.h"
#include "WMLGUIElement.h"

/**
 * Constructs a WMLTable object.
 *
 * @param c      the number of columns in the table
 * @param xb     the body of the table element in the parsetree
 * @param ww     the WMLWindow of the table
 */
WMLTable::WMLTable(int c, XMLBody *xb, WMLWindow *ww) {

  window = ww;
  windowInterface = ww->getWindowDrawingInterface();
  height = 0;
  columns = c;
  columnWidths = new int[columns];
  content = xb;

  int i;
  for(i = 0; i < columns; i++)
    columnWidths[i] = 0;

  XMLBody *xb1 = content, *xb2;
  XMLBody *tr, *td;
  int column, cellWidth, cellHeight;
  
  for(tr = xb1->getElement("tr"); tr != 0; tr = xb1->getElement("tr")) {

    xb2 = tr->getElement()->getBody();

    column = 0;
    currentRowHeight = 0;

    for(td = xb2->getElement("td"); td != 0; td = xb2->getElement("td")) {
      /* td is now the pointer in the parsetree which directs into the
         current cell */

      if(column < columns) {
        /* only do someting when the wml data is within the width 
           of the table */

        getCellSize(td->getElement()->getBody(), &cellWidth, &cellHeight);

        if(cellHeight > currentRowHeight)     currentRowHeight = cellHeight;
        if(cellWidth  > columnWidths[column]) columnWidths[column] = cellWidth; 

      }
      xb2 = td->getNextXMLBody();
      if(!xb2) break;
      column++;

    }

    height += currentRowHeight + 2;

    xb1 = tr->getNextXMLBody();
    if(!xb1) break;

  }
  
}

/**
 * Delete the object.
 */
WMLTable::~WMLTable() {

  delete columnWidths;

}

/**
 * Determines the cell size of the cell with the content specified by xb
 * @param xb         the body of the cell
 * @param cellWidth  when this function returns, this integer contains 
 * the cell width in pixels
 * @param cellHeight when this function returns, this integer contains 
 * the cell height in pixels
 */
void WMLTable::getCellSize(XMLBody *xb, int *cellWidth, int *cellHeight) {

  *cellWidth = 0;
  *cellHeight = 0;

  getCellSizeR(xb, cellWidth, cellHeight);

  return;

}

/* this must be different when rendering different styles, this will do for now */
void WMLTable::getCellSizeR(XMLBody *xb, int *cellWidth, int *cellHeight) {

  while(xb != 0) {

    if(xb->getContentType() == XMLBody::typeCharacterData) {

      if(windowInterface->getStringHeight(xb->getCharacterDataAsString(),0) > *cellHeight)
        *cellHeight = windowInterface->getStringHeight(xb->getCharacterDataAsString(),0);
      
      *cellWidth += windowInterface->getStringWidth(xb->getCharacterDataAsString(),0);

    } else {

      if(xb->getContentType() == XMLBody::typeElement) {

        if( xb->getElement()->getTagName()->equals("em")     || 
            xb->getElement()->getTagName()->equals("strong") ||
            xb->getElement()->getTagName()->equals("b")      ||
            xb->getElement()->getTagName()->equals("i")      ||
            xb->getElement()->getTagName()->equals("u")      ||
            xb->getElement()->getTagName()->equals("big")    ||
            xb->getElement()->getTagName()->equals("small")  ||
            xb->getElement()->getTagName()->equals("br")     ||
            xb->getElement()->getTagName()->equals("anchor") ||
            xb->getElement()->getTagName()->equals("a")        ) 
          getCellSizeR(xb->getElement()->getBody(), cellWidth, cellHeight);


        if(xb->getElement()->getTagName()->equals("img")) {

            WMLGUIElement *wbmp = 
              window->getImage(xb->getElement()->getAttributes()->getValue("src"));

            if(wbmp->getHeight() > *cellHeight)
              *cellHeight = wbmp->getHeight();
      
            *cellWidth += wbmp->getWidth();

        }
      }

    }

    xb = xb->getNextXMLBody();

  }

}

/**
 * Returns the width of the table.
 * @return the width of the table
 */
int WMLTable::getWidth() {
  int width = 0, i;

  for(i = 0; i < columns; i++)
    width += columnWidths[i] + 2;  /* add yo  all column widths */

  return width;

}

/**
 * Returns the height of the table.
 * @return the height of the table
 */
int WMLTable::getHeight() {

  return height;

}

/**
 * Renders the table.
 * @param x the x coordinate
 * @param y the y coordinate
 */
void WMLTable::display(int x, int y) {
  XMLBody *xb1 = content, *xb2, *tr, *td;
  int cellWidth, cellHeight, column;
  int xOffset, yOffset = 0;
  
  for(tr = xb1->getElement("tr"); tr != 0; tr = xb1->getElement("tr")) {

    xb2 = tr->getElement()->getBody();

    xOffset = 0;
    column = 0;

    currentRowHeight = 0;

    for(td = xb2->getElement("td"); td != 0; td = xb2->getElement("td")) {
      /* td is now the pointer in the parsetree which directs into the
         current cell */
      getCellSize(xb2->getElement()->getBody(), &cellWidth, &cellHeight);
      if(cellHeight > currentRowHeight) currentRowHeight = cellHeight;

      if(column < columns) {

        /* only do someting when the wml data is within the width 
           of the table */

        displayCell(xb2->getElement()->getBody(),x + xOffset,y + yOffset);
        xOffset += columnWidths[column] + 2;

      }

      column++;
      xb2 = td->getNextXMLBody();
      if(!xb2) break;
  
    }

    yOffset += currentRowHeight + 2;
    xb1 = tr->getNextXMLBody();
    if(!xb1) break;

  }
  
}

/**
 * Renders one cell.
 * @param xb the content to render
 * @param x x coordinate
 * @param y y coordinate
 * @return the width of the cell
 */
int WMLTable::displayCell(XMLBody *xb, int x, int y) {
  int tempx = x;

  while(xb != 0) {

    if(xb->getContentType() == XMLBody::typeCharacterData) {

      windowInterface->drawText(xb->getCharacterDataAsString(),x,y,0);
      x += windowInterface->getStringWidth(xb->getCharacterDataAsString(),0);

    } else {

      if(xb->getContentType() == XMLBody::typeElement) {

        if(xb->getElement()->getTagName()->equals("img")) {
          
          WMLGUIElement *wbmp = 
            window->getImage(xb->getElement()->getAttributes()->getValue("src"));
          wbmp->setXCoordinate(x);
          wbmp->setYCoordinate(y);
          windowInterface->drawImage(wbmp);

        } else
          x += displayCell(xb->getElement()->getBody(),x,y);
      }

    }

    xb = xb->getNextXMLBody();

  }

  return x - tempx;

}
