/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: WMLWindow.cpp,v 1.62 2001/03/23 11:49:34 wouter Exp $ 
 */

#include "WMLWindow.h"
#include "../wmlparser/XMLAttributeList.h"
#include "../wmlparser/XMLBody.h"
#include "../wmlparser/XMLCharData.h"
#include "../wmlparser/XMLElement.h"
#include "../wmlparser/WMLParser.h"
#include "TestWindowDrawingInterface.h"
#include "WBMP.h"
#include "WindowDrawingInterface.h"
#include "WMLGUIElement.h"
#include "WMLTable.h"

/**
 * Constructs a WMLWindow object, which is gonna use the 
 * WindowDrawingInterface wdi.
 * @param wdi the WindowDrawingInterface that must be used
 */
WMLWindow::WMLWindow(WindowDrawingInterface *wdi) {

  parseTree = NULL;
  currentCard = NULL;
  windowInterface = wdi;
  vScrollBar = false;
  hScrollBar = false;
  vScroll = 0;
  hScroll = 0;
  lastReadLinkTarget = NULL;
  topLinkTarget = NULL;
  windowWidth = 0;
  windowHeight = 0;
  sourceHeight = 0;
  handMousePointer = false;
  uri = NULL;
  history = new PointerList();
  historyIndex = -1;

}

/**
 * Delete the object
 */
WMLWindow::~WMLWindow() {

  while(history->getSize() != 0)
    delete (URI *) history->removeFirstAdded();

  delete history;
  delete parseTree;
  delete uri;
  clearCardDependentMemory();
  
}

/**
 * This function tests the code of this class
 *
 * @return true when all tests succeed, otherwise false
 */
bool WMLWindow::test() {

  TestWindowDrawingInterface twdi;
  WMLWindow ww ((WindowDrawingInterface *) &twdi);
  
  URI uri("http://www.catchy.net/wap/index.wml");

  ww.setSource(&uri);

  ww.setWindowDimensions(12,12);
  ww.drawContentWindow();

  return true;

}

/**
 * Returns the uri of the current document
 * @return the uri
 */
URI *WMLWindow::getURI() {

  return uri;

}

/**
 * Returns the content sink
 * @return the content sink
 */
ContentSink *WMLWindow::getContentSink() {

  return &contentSink;

}

/**
 * Returns the window drawing interface.
 * @return the window drawing interface
 */
WindowDrawingInterface *WMLWindow::getWindowDrawingInterface() {

  return windowInterface;

}

/**
 * Returns information on the shape of the mouse pointer
 * @return true when mouse pointer must be hand-shaped, false for a
 *         normal mouse shape
 */
bool WMLWindow::getMousePointerShape() {

  return handMousePointer;

}

/** 
 * Sets the uri of the source.
 *
 * @param u the uri of the source
 */
void WMLWindow::setSource(URI *u) {

  setSource(u, true);

}

/**
 * Internal version of setSource().
 * @param u the URI to change to.
 * @param updateHistory whether or not this change should be
 * recorded in the history.
 */
void WMLWindow::setSource(URI *u, bool updateHistory) {
  char *buffer;
  size_t length;
  String s;

  if(updateHistory)
    historyAdd(u);

  uri = new URI(u->toString());

  clearCardDependentMemory();

  updater.add(UpdateWindowList::actionClearButtons, NULL);
  /* clear the buttons */

  if(parseTree != NULL) {

    delete parseTree;   /* remove old parsetree */
    parseTree = NULL;
    currentCard = NULL;

  }

  ContentSinkResult *csr = contentSink.get(uri, NULL);

  csr->getContent((void **) &buffer, &length);

  String source(buffer, length);

  delete csr;

  WMLParser parser(&source); /* parse the new source */

  if(parser.getErrorMessage()) {

    String s("<wml>\n<card>\n<p>ERROR</p></card>\n</wml>\n");

    WMLParser errorParser(&s);

    String* charData = errorParser.getRootElement()->getBody()->
      getElement("card")->getElement()->getBody()->
      getElement("p")->getElement()->getBody()->getCharacterDataAsString();

    charData->clear(); /* remove 'ERROR' */
    charData->append(parser.getErrorMessage()); /* append real error text */

    parseTree = new XMLElement(errorParser.getRootElement());

  } else {

    /* copy the parsetree */
    parseTree = new XMLElement(parser.getRootElement());

  }

  String card = new String(uri->getFragment());
  setCard(&card);

  setButtonLabels();
  
}

/**
 * Sets the current card to the card in the argument.
 * @param s the id of the card to display
 */
void WMLWindow::setCard(String *s) {

  if(s->getLength() == 0) {

    currentCard = parseTree->getBody()->getElement("card")->getElement();

  } else {

    XMLBody *xb = parseTree->getBody()->getElement("card");

    while(xb != NULL) {

      if(xb->getElement()->getAttributes()->getValue("id")) {

        if(xb->getElement()->getAttributes()->getValue("id")->equals(s)) {
  
          currentCard = xb->getElement();
          updater.add(UpdateWindowList::actionClearButtons, NULL);
          break;

        }

      }

      xb = xb->getNextXMLBody();

    }

  }

  String location(uri->toString());
  updater.add(UpdateWindowList::actionURIChanged, &location);

}

/**
 * Sets the output window text. Makes sure that no double information
 * gets through the UpdateWindowList.
 * @param s  the desired text for the output window
 */
void WMLWindow::setOutputWindowText(String* s) {

  if(s == NULL) {
    if(outputWindowText.getLength() != 0) {

      outputWindowText.clear();
      updater.add(UpdateWindowList::actionClearOutputWindowText, NULL);

    }
  }
  else {
    if(!s->equals(&outputWindowText)) {

      outputWindowText.clear();
      outputWindowText.append(s);
      updater.add(UpdateWindowList::actionSetOutputWindowText, &outputWindowText);

    }
  }

}

/**
 * Clears the memory that is used to store  a single card. These are
 * the list of images, the list of input elements and the list of 
 * select elements.
 */
void WMLWindow::clearCardDependentMemory() {

  while(imageList.getSize() != 0)
    delete (WMLGUIElement *) imageList.removeFirstAdded();

  while(inputList.getSize() != 0)
    delete (WMLGUIElement *) inputList.removeFirstAdded();

  while(selectList.getSize() != 0)
    delete (WMLGUIElement *) selectList.removeFirstAdded();

  topLinkTarget = NULL;
  lastReadLinkTarget = NULL;
  windowHeight = 0;
  vScroll = 0;
  hScroll = 0;  /* clear all 'leftovers' from previous card */

}

/** 
 * Tells the Window that the mouse moved to location x,y.
 *
 * @param x    the x coordinate of the mouse pointer
 * @param y    the y coordinate of the mouse pointer
 */
void WMLWindow::mouseMove(int x, int y) {

  if(anchorList.getLinkTarget(x,y)) {

    if(!handMousePointer) {

      handMousePointer = true;
      updater.add(UpdateWindowList::actionSetHandMousePointer,NULL);      

    }
    setOutputWindowText(anchorList.getLinkTarget(x,y));

  } else {
    
    if(handMousePointer) {

      handMousePointer = false;      
      updater.add(UpdateWindowList::actionSetNormalMousePointer,NULL);
      setOutputWindowText(NULL);

    }
  }
}

/** 
 * Tells the Window that the mouse button was clicked on location x,y.
 *
 * @param x    the x coordinate of the mouse pointer
 * @param y    the y coordinate of the mouse pointer
 * @return true if the click caused the WMLWindow to follow a link.
 */
bool WMLWindow::mouseClick(int x, int y) {
  unsigned int i;
  bool hit = false;

  if(anchorList.getLinkTarget(x,y)) {

    hit = true;

    if(historyIndex != -1) {

      /* Throw away everything after the current history item. */
      for(i = historyIndex; i < history->getSize() - 1;) {
        
        delete (URI *) history->getLastAdded();
        history->removeLastAdded();

      }

    }

    updater.add(UpdateWindowList::actionSetNormalMousePointer,NULL);

    URI linkUri(anchorList.getLinkTarget(x,y)->getPtr());

    /* remove fragment from original uri, workaround for some URI bug.... */
    /* uri bug: URI::resolve("http://somewhere/file.wml#qwe","#asd") 
       does not work */
    /* the following 'if' block could be deleted when the uri bug is fixed */

    if(uri->getFragment() != NULL) {
      
      String temp(uri->toString());
      while (temp.getLastCharacter() != '#')
        temp.removeLastCharacter();

      temp.removeLastCharacter();

      delete uri;
      uri = new URI(temp.getPtr());

    }

    URI *resolvedURI = URI::resolve(getURI(),&linkUri);

    setSource(resolvedURI);

  }
  
  return hit;

}

/**
 * Returns the pointer in the parsetree in which the target of the 
 * link is denoted.
 * 
 * @param  xe the < a > or < anchor > element
 * @return a pointer from the parsetree with the target of the anchor
 */
String *WMLWindow::getLinkTarget(XMLElement *xe) {
  
  if(xe->getTagName()->equals("a")) 
    return xe->getAttributes()->getValue("href");

  if(xe->getTagName()->equals("anchor")) {

    if(xe->getBody()->getElement("go")) {

      xe = xe->getBody()->getElement("go")->getElement();
      if(!xe)
        return NULL;

      return xe->getAttributes()->getValue("href");

    }
  }
  return NULL;
 
}

/**
 * Tells the Window that the user pressed the button identified by
 * the argument.
 *
 * @param b    the button which was pressed
 */
void WMLWindow::pushedButton(button b) {

  /*if(b == buttonOK && topLinkTarget) 
    updater.add(UpdateWindowList::actionChangeSource, topLinkTarget);*/

}

/** 
 * Returns a list of things the window drawing interface must do to
 * the window.
 *
 * @return the list
 */
UpdateWindowList* WMLWindow::getUpdates() {

  return &updater;

}

/** 
 * Sets the labels according to the current card
 */
void WMLWindow::setButtonLabels() {

  setButtonLabels(currentCard->getBody());

}

/** 
 * Sets the labels according to the xml body in the argument which
 * must be the body of a 'card' element
 * 
 * @param xb the XMLBody from which the button labels must be acquired
 */
void WMLWindow::setButtonLabels(XMLBody *xb) {
  String usedNames;

  for(;xb != NULL;xb = xb->getNextXMLBody()) {
    
    if(xb->getContentType() == XMLBody::typeElement) {
  
      if(xb->getElement()->getTagName()->equals("do")) {

        /* check if the name of the do element is already used: */
        if(!usedNames.substring(xb->getElement()->getAttributes()->getValue("name"))) {

          /* check if the body is NOT a noop operation */
          if(!xb->getElement()->getBody()->getElement()->getTagName()->equals("noop")) {

            /* check if the do element has type 'prev':*/
            /* (type is a required attribute of do) */
            if(xb->getElement()->getAttributes()->getValue("type")->equals("prev")) {

              String label(xb->getElement()->getAttributes()->getValue("label"));

              if(label.getLength() == 0) {

                /* the wml writer did not specify a label */
                label.append("Back");

              }              
              updater.add(UpdateWindowList::actionChangeBackButtonLabel,&label);

            } else {

              /* the type of the do element is not 'prev' */
              addButtonLabel(xb);

            }
          }

          if(xb->getElement()->getAttributes()->getValue("name")) {
          
            /* add the name attribute to the usedName list */
            usedNames.append(xb->getElement()->getAttributes()->getValue("name")); 
            usedNames.append("\1");

          }
        }
      } 
      setButtonLabels(xb->getElement()->getBody());
    }   
  }
}

/** 
 * Sets the label of a button  according to the xml body in the 
 * argument which getElement()->getTagName() must be 'do'.
 * 
 * @param xb the XMLBody from which the button label must be acquired
 */
void WMLWindow::addButtonLabel(XMLBody *xb) {

  /* when the body consists of a noop operation do nothing:*/
  if(xb->getElement()->getBody()->getElement()->getTagName()->equals("noop")) 
    return;

  String s(xb->getElement()->getAttributes()->getValue("label"));

  if(s.getLength() == 0) {
    /* the wml writer did not specify a label */
    s.append("Button");
  }
    
  updater.add(UpdateWindowList::actionAddButtonLabel,&s);

}

/** 
 * Displays the content according to the source in the content window,
 * with use of the WindowDrawingInterface. For displaying content do
 * the following:
 * \code
 *  // here: determine size of client area, put dimensions in x and y 
 *  ww->setWindowDimensions(x,y);
 *  while(ww->getUpdates()->getFirstAction() != actionNoMoreActions) {
 *  
 *   // here: do the updates
 *   // here: determine size of client area, put dimensions in x and y 
 *   ww->setWindowDimensions(x,y);
 *
 *  }
 *  ww->drawContentWindow();
 * \endcode
 */
void WMLWindow::drawContentWindow() {

  drawContentWindow(true);

}

/** 
 * Displays the content according to the source in the content window,
 * with use of the WindowDrawingInterface. When draw=false this member
 * is only used to determine the  height and width of the source, so
 * no actual drawing will happen. 
 *
 * @param draw when false, does not really draw, determines only the 
 * size of the content
 */
void WMLWindow::drawContentWindow(bool draw) {

  /* clear all lists: */
  styleList.clear();
  anchorList.clear();
  while(bodyList.getSize())
    bodyList.removeLastAdded();

  offsetList.clear();
  
  readDisplayedLinks = 0;
  readPeerElements = 0; 
  readImages = 0;
  sourceHeight = 0;

  if(currentCard == NULL) {

    windowInterface->startDrawing(windowWidth, windowHeight);
    windowInterface->stopDrawing();
    return;

  }

  if(draw)
    determineSize = false;
  else
    determineSize = true;

  /* determineSize means that the renderer is only going through the 
     the parse tree to determine the size of the source, so no calls
     to drawText, etc. will be made */

  if(!determineSize)
    windowInterface->startDrawing(windowWidth, windowHeight);
  else
    windowInterface->startDrawing(0, 0);

  XMLBody *xb;

  /* walk throug the currentCard and search for p or pre elements: */
  for(xb = currentCard->getBody(); xb != NULL; xb = xb->getNextXMLBody()) {

    if(xb->getElement()->getTagName()->equals("p") || 
       xb->getElement()->getTagName()->equals("pre") ) {
      /* the <p> and <pre> tag are treated the same for now */

      wrap = true;
      currentAlignmentMethod = alignLeft;

      /* determine the wrap mode for the current paragraph, the default
         value is 'wrap', so only change the boolean wrap when the
         attribute is present and set to 'nowrap' */
      String *value = xb->getElement()->getAttributes()->getValue("mode");
      if(value && value->equals("nowrap")) 
        wrap = false;

      /* determine the allign method for the current pararaph, the
         default value is 'left' */
      value = xb->getElement()->getAttributes()->getValue("align");
      if(value && value->equals("center")) 
        currentAlignmentMethod = alignCenter;
      if(value && value->equals("right")) 
        currentAlignmentMethod = alignRight;

      /* render the paragraph */
      styleList.add(styleNoStyle); 
      displayParagraph(xb->getElement());

    }

  }
  windowInterface->stopDrawing();
  sourceHeight += 3;

}

/** 
 * Displays a wml paragraph.
 *
 * @param xe the paragraph element
 */
void WMLWindow::displayParagraph(XMLElement *xe) {

  breakPossible = false;

  /* the currentLocation goes from the upper left corner to the
     lower goes corner while the algoritm goes into the parsetree */
  currentLocation.x = 0;
  currentLocation.y = sourceHeight;

  currentLineHeight = 0; /* this variable is used to determine the 
                            amount of pixels that must be skipped
                            verically when a < br/ > occurs */
 
  displayParagraph(xe->getBody());

}

/** 
 * Displays a wml paragraph. It calls itself recursively when it has
 * read an element. The currentLocation variable is the point at which
 * the algorithm is at. It first goes from left to right, top-down 
 * till it is at the end of the paragraph and then it's going from
 * right to left, bottom-up and is used to determine where to 
 * display the content. When the algorithm is going top-down it 
 * determines where newlines must occur. The content is not displayed
 * until the algorithm is going bottom up for vertical alignment
 * purposes.
 *
 * @param xb  the body element that must be displayed
 */
void WMLWindow::displayParagraph(XMLBody *xb) {

  if(xb == NULL || xb->getContentType() == XMLBody::typeNothing) {

    styleList.removeLastAdded();
    /* every chain of XMLBody elements has a certain style, this
       chain is ended now, so the last added style must be removed */

    if(bodyList.getSize() == 0) {

      currentLocation.y += currentLineHeight;
      
      /* update the source height, the next paragraph 
         continues from sourceHeight further down*/
      sourceHeight = currentLocation.y;
      currentLineHeight = 0;
      newLineForward();
      newLineBackward();
      currentLocation.y--;
      currentLineHeight = endOfParagraph;
      /* this constant is put in currentLineHeight so that when a 
         < br > was put at the end of the paragraph this < br > is
         ignored */

    }
    else {

      /* this chain of XMLBody elements is ended, go further with the
         next one */

      XMLBody *xb2 = (XMLBody *)bodyList.removeLastAdded();
      displayParagraph(xb2);

    }
    return;

  }

  switch(xb->getContentType()) {

  case (XMLBody::typeCharacterData):
    /* add the next xmlbody to the bodylist, this will be handled
       later:*/
    bodyList.add(xb->getNextXMLBody());
    
    {
      XMLCharData *xcd = xb->getCharacterData();
      xcd->rewind();
      displayParagraphText(xcd); /* render char data */
    }
    break;
    
  case (XMLBody::typeElement):
 
    if(xb->getElement()->getTagName()->equals("br")) {

      if(currentLineHeight == 0)
        currentLineHeight = 11;

      newLineForward();
      displayParagraph(xb->getNextXMLBody());
      
      if(currentLineHeight == 0)
        currentLineHeight = 11;

      if(currentLineHeight == endOfParagraph)
        currentLineHeight = 0;

      newLineBackward();

      return;

    }

    if(xb->getElement()->getTagName()->equals("i")) {

      /* put the next xmlbody, the first content that must NOT
         be rendered italic, in the list */
      bodyList.add(xb->getNextXMLBody());

      /* add the italic style to the style list */
      styleList.add(styleItalic);

      /* start rendering the xb->getElement()->getBody() chain 
         italic, this function continues with the chains in
         the bodylist when that chain is done */
      displayParagraph(xb->getElement()->getBody());
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("b")) {

      /* likewise, see italic */
      bodyList.add(xb->getNextXMLBody());
      styleList.add(styleBold);
      displayParagraph(xb->getElement()->getBody());
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("big")) {

      bodyList.add(xb->getNextXMLBody());
      styleList.add(styleBig);
      displayParagraph(xb->getElement()->getBody());
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("small")) {

      bodyList.add(xb->getNextXMLBody());
      styleList.add(styleSmall);
      displayParagraph(xb->getElement()->getBody());
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("u")) {

      bodyList.add(xb->getNextXMLBody());
      styleList.add(styleUnderlined);
      displayParagraph(xb->getElement()->getBody());
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("em")) {

      bodyList.add(xb->getNextXMLBody());
      styleList.add(styleEm);
      displayParagraph(xb->getElement()->getBody());
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("strong")) {

      bodyList.add(xb->getNextXMLBody());
      styleList.add(styleStrong);
      displayParagraph(xb->getElement()->getBody());
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("fieldset")) {

      bodyList.add(xb->getNextXMLBody());
      styleList.add(0);
      displayParagraph(xb->getElement()->getBody());
      return;
      
    }


    if(xb->getElement()->getTagName()->equals("anchor") ||
       xb->getElement()->getTagName()->equals("a")) {

      bodyList.add(xb->getNextXMLBody());
      styleList.add(styleLink);
      lastReadLinkTarget = getLinkTarget(xb->getElement());
      displayParagraph(xb->getElement()->getBody());
      return;

    }

    if(xb->getElement()->getTagName()->equals("input")) {

      displayInputElement(xb);
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("select")) {

      displaySelectElement(xb);
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("table")) {

      displayTableElement(xb);
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("img")) {

      displayImageElement(xb);
      return;
      
    }

    if(xb->getElement()->getTagName()->equals("go")       ||
       xb->getElement()->getTagName()->equals("prev")     ||
       xb->getElement()->getTagName()->equals("refresh")  ||
       xb->getElement()->getTagName()->equals("do")) 

      displayParagraph(xb->getNextXMLBody());
    
    break;

  default:
    break;

  }
}

/**
 * Displays an select element
 * @param xb  the body element that must be displayed
 */
void WMLWindow::displaySelectElement(XMLBody *xb) {
  bool newLine = false;
  WMLGUIElement *wse;
  
  if((wse = selectList.getElement(readPeerElements)) == (void *)selectList.notFound) {
    PointerList optionList;

    getOptions(&optionList,xb->getElement()->getBody());
    wse = new WMLGUIElement(windowInterface->createSelectElement(&optionList), 
                            readPeerElements);
      
    /* the objects pointed to in the pointer list are deleted
       when clearCardDependentMemory() is called */

    selectList.add(wse);

  }

  readPeerElements++;

  int width = wse->getWidth();
  int height = wse->getHeight();

  if(currentLocation.x + width > sourceWidth) {
    /* the select element doesn't fit on the line */
    if(wrap && breakPossible) {

      newLineForward();
      newLine = true;

    } else 
      sourceWidth = currentLocation.x + width;

  }

  updateCurrentLineHeight(height);
  updateSourceWidth(currentLocation.x);

  currentLocation.x += width;
  breakPossible = true;

  displayParagraph(xb->getNextXMLBody());

  /* the size is now determined, nothing has to be done when
     determineSize == true */
  if(determineSize) 
    return;

  currentLocation.x -= width;

  updateCurrentLineHeight(height);

  if(wse->getPeer()) {

    wse->setXCoordinate(currentLocation.x - hScroll);
    wse->setYCoordinate(currentLocation.y - vScroll - height + 
                        windowInterface->getPixelsBelowBaseLine(styleList.getStyle()));

    windowInterface->drawSelectElement(wse);

  }

  if(newLine) 
    newLineBackward();

}

/**
 * Puts the options from the parsetree into a list
 *
 * @param wol the PointerList that must be filled with the options
 * @param xb  the body of the select element in the parsetree
 */
void WMLWindow::getOptions(PointerList* wol, XMLBody* xb) {

  if(xb == NULL) 
    return;

  if(xb->getContentType() == XMLBody::typeElement) {

    if(xb->getElement()->getTagName()->equals("optgroup"))
      getOptions(wol,xb->getElement()->getBody());

    if(xb->getElement()->getTagName()->equals("option"))
      getOptions(wol,xb->getElement()->getBody());

    /* when tagname == "onevent": do nothing */

  } 

  if(xb->getContentType() == XMLBody::typeCharacterData) 
    wol->add(new String(xb->getCharacterDataAsString()));

  getOptions(wol,xb->getNextXMLBody());

}

/**
 * Displays an input element
 * @param xb              the body element that must be displayed
 */
void WMLWindow::displayInputElement(XMLBody* xb) {
  bool newLine = false;
  WMLGUIElement *wie;
  
  if((wie = inputList.getElement(readPeerElements)) == (void *) inputList.notFound) {

    wie = new WMLGUIElement(windowInterface->createInputElement(), readPeerElements);
    inputList.add(wie);

    /* the objects pointed to in the pointer list are deleted
       when clearCardDependentMemory() is called */

  }

  readPeerElements++;

  int width = wie->getWidth();
  int height = wie->getHeight();
 
  if(currentLocation.x + width > sourceWidth) {
    /* the input element doesn't fit on the line */
    if(wrap && breakPossible) {

      newLineForward();
      newLine = true;

    } else 
      sourceWidth = currentLocation.x + width;

  }

  updateCurrentLineHeight(height);
  updateSourceWidth(currentLocation.x);

  currentLocation.x += width;
  breakPossible = true;


  displayParagraph(xb->getNextXMLBody());

  /* the size is now determined, nothing has to be done when
     determineSize == true */
  if(determineSize) 
    return;

  currentLocation.x -= width;

  updateCurrentLineHeight(height);

  if(wie) {

    wie->setXCoordinate(currentLocation.x - hScroll);
    wie->setYCoordinate(currentLocation.y - vScroll - height + windowInterface->getPixelsBelowBaseLine(styleList.getStyle()));

    windowInterface->drawInputElement(wie);

  }

  if(newLine) 
    newLineBackward();

}

/**
 * Displays an img element
 * @param xb  the xml body element that is the img element
 */
void WMLWindow::displayImageElement(XMLBody *xb) {
  bool newLine = false;
  WMLGUIElement *wbmp = getImage(xb->getElement()->getAttributes()->getValue("src"));

  int width = wbmp->getWidth();
  int height = wbmp->getHeight();

  if(currentLocation.x + width > sourceWidth) {
    /* the select element doesn't fit on the line */
    if(wrap && breakPossible) {

      newLineForward();
      newLine = true;

    } else 
      sourceWidth = currentLocation.x + width;

  }

  updateCurrentLineHeight(height);
  updateSourceWidth(currentLocation.x);

  currentLocation.x += width;
  breakPossible = true;

  displayParagraph(xb->getNextXMLBody());

  /* the size is now determined, nothing has to be done when
     determineSize == true */
  if(determineSize) 
    return;

  currentLocation.x -= width;

  updateCurrentLineHeight(height);

  if(wbmp->getPeer()) {

    wbmp->setXCoordinate(currentLocation.x - hScroll);
    wbmp->setYCoordinate(currentLocation.y - vScroll - height + 
                         windowInterface->getPixelsBelowBaseLine(styleList.getStyle()));

    windowInterface->drawImage(wbmp);

  }

  if(newLine) 
    newLineBackward();

}

/**
 * Gets the image that is about to be rendered from the image list.
 * When it is not yet in the list it is put in the list by the WBMP
 * class. 
 * @param src the src attribute of the img element.
 * @return the WMLGUIELement pointer from the list
 */
WMLGUIElement *WMLWindow::getImage(String *src) {

  WMLGUIElement *wbmp;

  if((wbmp = imageList.getElement(readImages)) == (void *)imageList.notFound) {

    WBMP image(this,src);

    wbmp = new WMLGUIElement(windowInterface->createImage(&image), readImages);
      
    /* the objects pointed to in the pointer list are deleted
       when clearCardDependentMemory() is called */

    imageList.add(wbmp);

  }
  readImages++;

  return wbmp;

}

/**
 * Moves in the history. If this call returns true, the move
 * was made and the display needs to be refreshed.
 * @param direction integer specifying how much to move (e.g. 1 is 
 * move 1 page forward, -1 is move one page backward).
 * @return true if the move was made (otherwise the direction was out of
 * bounds).
 */
bool WMLWindow::historyGo(int direction) {
  int newIndex = historyIndex + direction;
  bool r = false;
  String s;

  if(newIndex >= 0 && newIndex < (signed int) history->getSize()) {

    r = true;

    historyIndex = newIndex;
    setSource(((URI **) (history->getArray()))[historyIndex], false);

    /* Properly enable / disable back & forward buttons. */
    s.append((history->getSize() > 1 && historyIndex > 0) ? "on" : "off");
    updater.add(UpdateWindowList::actionBackButton, &s);

    s.clear();

    s.append(((signed int) history->getSize() > 1 && 
              historyIndex < (signed int) history->getSize() - 1) ? 
             "on" : "off");
    updater.add(UpdateWindowList::actionForwardButton, &s);

  }

  return r;

}

/**
 * Displays a table element
 * @param xb  the xml body element that is the table
 */
void WMLWindow::displayTableElement(XMLBody *xb) {
  bool newLine = false;
  
  WMLTable table(
                 xb->getElement()->getAttributes()->getValue("columns")->getInt(),
                 xb->getElement()->getBody(),
                 this);

  int width = table.getWidth();
  int height = table.getHeight();
  
  if(currentLocation.x + width > sourceWidth) {
    /* the table element doesn't fit on the line */
    if(wrap && breakPossible) {

      newLineForward();
      newLine = true;

    } else 
      sourceWidth = currentLocation.x + width;

  }

  updateCurrentLineHeight(height);
  updateSourceWidth(currentLocation.x);

  currentLocation.x += width;
  breakPossible = false;

  displayParagraph(xb->getNextXMLBody());

  /* the size is now determined, nothing has to be done when
     determineSize == true */
  if(determineSize) 
    return;

  currentLocation.x -= width;

  updateCurrentLineHeight(height);

  table.display(currentLocation.x - hScroll,
                currentLocation.y - vScroll 
                - height
                + windowInterface->getPixelsBelowBaseLine(styleList.getStyle()));

  if(newLine) 
    newLineBackward();

}

/** 
 * Displays some wml character data. It calls itself recursively when
 * it has read the char data. The currentLocation variable is the point
 * at which the algorithm is at. It first goes from right to left, till
 * it is at the end of the paragraph and then it's going from left to 
 * right and it is used to determine where to display the content. 
 * Notice that this function calls itself from the middle of the function.
 *
 * @param xcd  the chardata that must be displayed, per function call 
 *             this function only reads one word, or part of a word
 */
void WMLWindow::displayParagraphText(XMLCharData *xcd) {
  String s;
  xcd->getNextWord(&s);

  if(s.getLength() == 0) {

    /* the end of the char data is reached, go further with the
       next element in the bodyList */

    XMLBody *xb = (XMLBody *)bodyList.removeLastAdded();
    displayParagraph(xb);

    return;

  }

  bool newLine = false;
  int style = styleList.getStyle();
  String *linkTarget = lastReadLinkTarget;
  int height = windowInterface->getStringHeight(&s,style);
  int stringWidth = windowInterface->getStringWidth(&s, style);

  if(!s.equals(" ")) {

    StyleList sl(&styleList);
    PointerList xbl(&bodyList);
    int wordWidth = stringWidth + additionalWordLength(xcd,&xbl,&sl);

    if(currentLocation.x + wordWidth > sourceWidth) {
      /* the word doesn't fit on the line */
      if(wrap && breakPossible) {

        newLineForward();
        newLine = true;

      } else 
        sourceWidth = currentLocation.x + stringWidth;
    }

  }

  updateCurrentLineHeight(height);

  if(style & styleLink) {

    if(currentLocation.y - vScroll + height / 2 > 0) {

      /* the link is in the display are of the window */

      readDisplayedLinks++;
    
      if(readDisplayedLinks == 1 || linkTarget == topLinkTarget) {

        /* (when linkTarget == topLinkTarget this word is part 
           of the link and so this word must also be 
           rendered as a top link) */

        topLinkTarget = linkTarget;
        style = style + styleToplink;

      }
    }
  }

  /* do not update currentLocation.x when there is a space at the
     beginning of a line */
  if(!(currentLocation.x == 0 && s.equals(" "))) 
    currentLocation.x += stringWidth;

  if(!s.equals(" "))
    updateSourceWidth(currentLocation.x);

  if(s.equals(" ")) {

    if(currentLocation.x == 0)
      breakPossible = false; /* a break is not possible when for a 
                                word on the first line */
    else
      breakPossible = true;

  } else 
    breakPossible = false;

  displayParagraphText(xcd);

  /* the size is now determined, nothing has to be done when
     determineSize == true */
  if(determineSize) 
    return;

  if(!(currentAlignmentMethod == alignRight &&
       currentLocation.x == sourceWidth &&
       s.equals(" ")))
    currentLocation.x -= stringWidth;

  updateCurrentLineHeight(height);

  int x = currentLocation.x - hScroll;
  int y = currentLocation.y - vScroll - height + windowInterface->getPixelsBelowBaseLine(style);

  if(y <= windowHeight && y + height > 0) {

    /* only display when the word is in the visable area */

    if(style & styleLink)
      anchorList.add(x,y,x + stringWidth,y + height, linkTarget);

    if(!s.equals(" ") || style & styleUnderlined || style & styleLink)
      windowInterface->drawText(&s, x, y, style);
  
  }

  if(newLine)    
    newLineBackward();

}

/** 
 * The render algorithm first goes top-down and later bottom-up in the
 * parsetree. When it goes top-down and must let a newline occurs it
 * calls newLineForward, when later it goes bottom-up it calls 
 * newLineBackward. This function udpates the currentLocation,
 * currentLineHeight and breakPossible variables.
 */
void WMLWindow::newLineForward() {

  switch(currentAlignmentMethod) {

  case alignLeft:
    offsetList.add(currentLocation.x);
    break;

  case alignCenter:
    offsetList.add(currentLocation.x);
    break;
  
  case alignRight:
    break;
  }
  
  currentLocation.y += currentLineHeight;
  currentLocation.x = 0;
  currentLineHeight = 0;
  breakPossible = false;

}

/** 
 * The render algorithm first goes top-down and later bottom-up in the
 * parsetree. When it goes top-down and must let a newline occurs it
 * calls newLineForward, when later it goes bottom-up it calls 
 * newLineBackward. This function udpates the currentLocation,
 * variable.
 */
void WMLWindow::newLineBackward() {

  currentLocation.y -= currentLineHeight;
  currentLineHeight = 0;
  
  switch(currentAlignmentMethod) {

  case alignLeft:
    currentLocation.x = offsetList.getLastAdded();
    offsetList.removeLastAdded();
    break;

  case alignCenter:
    currentLocation.x = (sourceWidth + offsetList.getLastAdded()) / 2;
    offsetList.removeLastAdded();
    break;

  case alignRight:
    currentLocation.x = sourceWidth;
    break;

  }
}


/** 
 * Updates the currentLineHeight object variable when the integer in the
 * argument is heigher than the currentLineHeight;
 * @param height  the new height when height > currentLineHeight
 */
void WMLWindow::updateCurrentLineHeight(int height) {

  if(height > currentLineHeight)
    currentLineHeight = height;

}

/**
 * Updates the sourceWidth object variable when the integer in the
 * argument is bigger.
 * @param x  the new sourceWidth when x > sourceWidth
 */
void WMLWindow::updateSourceWidth(int x) {

  if(x > sourceWidth) 
    sourceWidth = x;

}

/** 
 * Tells the wmlwindow that the user wants to scroll i pixels down.
 * Use in the following manner:
 * \code
 *  ww->scrollVertical(3);
 *  if(ww->getUpdates()->getFirstAction() != 
 *               UpdateWindowList::actionNoMoreActions) {
 *     // draw the window again
 *
 *  }
 * \endcode
 *
 * @param  i  the amount of pixels to scroll down
 */
void WMLWindow::scrollVertical(int i) {

  if(!vScrollBar) 
    return;
  int oldvScroll = vScroll;
  vScroll += i;
  if(vScroll < 0) 
    vScroll = 0;
  if(vScroll > sourceHeight - windowHeight) 
    vScroll = sourceHeight - windowHeight;
  if(oldvScroll == vScroll) 
    return;
  
  String s;
  s.append(vScroll);
  updater.add(UpdateWindowList::actionVerticalScrollBarPos,&s);
  
}

/** 
 * Tells the wmlwindow that the user wants to scroll vertical to
 * position i.  Use in the following manner:
 * \code
 *  ww->setVertical(3);
 *  if(ww->getUpdates()->getFirstAction() != 
 *               UpdateWindowList::actionNoMoreActions) {
 *     // draw the window again
 *
 *  }
 * \endcode
 *
 * @param  i  the amount of pixels that are scrolled down
 */
void WMLWindow::setVertical(int i) {

  scrollVertical(i - vScroll);
  
}

/** 
 * Tells the wmlwindow that the user wants to scroll i pixels to the
 * right. Use in the following manner:
 * \code
 *  ww->scrollHorizontal(3);
 *  if(ww->getUpdates()->getFirstAction() != 
 *               UpdateWindowList::actionNoMoreActions) {
 *
 *     // draw the window again
 *
 *  }
 * \endcode
 *
 * @param  i  the amount of pixels to scroll to the right
 */
void WMLWindow::scrollHorizontal(int i) {

  if(!hScrollBar) 
    return;

  int oldhScroll = hScroll;
  
  hScroll += i;
  
  if(hScroll < 0)
    hScroll = 0;

  if(hScroll > sourceWidth - windowWidth)
    hScroll = sourceWidth - windowWidth;

  if(oldhScroll == hScroll)
    return;
  
  String s;
  s.append(hScroll);
  updater.add(UpdateWindowList::actionHorizontalScrollBarPos,&s);

}

/** 
 * Tells the wmlwindow that the user wants to scroll i pixels to the
 * right. Use in the following manner:
 * \code
 *  ww->setHorizontal(3);
 *  if(ww->getUpdates()->getFirstAction() != 
 *               UpdateWindowList::actionNoMoreActions) {
 *
 *     // draw the window again
 *
 *  }
 * \endcode
 *
 * @param  i  the amount of pixels that must be scrolled to the right
 */
void WMLWindow::setHorizontal(int i) {

  scrollHorizontal(i - hScroll);

}

/** 
 * Sets the window dimensions. Use this function before displaying
 * the window:
 * \code
 *  // here: determine size of client area, put dimensions in x and y 
 *  ww->setWindowDimensions(x,y);
 *  while(ww->getUpdates()->getFirstAction() != 
 *               UpdateWindowList::actionNoMoreActions) {
 *  
 *   // here: do the updates
 *   // here: determine size of client area, put dimensions in x and y 
 *   ww->setWindowDimensions(x,y);
 *
 *  }
 *  ww->drawContentWindow();
 * \endcode
 *
 * @param  width  the width of the client area of the window in pixels
 * @param  height the height of the client area of the window in pixels
 */
void WMLWindow::setWindowDimensions(int width, int height) {
  double vs = 0;
  double hs = 0;

  if(windowWidth != width || windowHeight != height) {

    windowWidth = width;
    windowHeight = height;

    if(sourceHeight != 0)
      vs = (double)vScroll / (double)sourceHeight;

    if(sourceWidth != 0)
      hs = (double)hScroll / (double)sourceWidth;

    sourceWidth = windowWidth;

    drawContentWindow(false);

    if(sourceWidth > windowWidth)
      drawContentWindow(false);

    String s;

    if(sourceWidth > windowWidth) {

      if(hScrollBar == false) {

        s.append("on");
        updater.add(UpdateWindowList::actionHorizontalScrollBar,&s);
        s.clear();
        hScrollBar = true;

      }

      s.append(sourceWidth);
      updater.add(UpdateWindowList::actionHorizontalScrollBarRange,&s);
      s.clear();
      s.append(windowWidth);
      updater.add(UpdateWindowList::actionHorizontalScrollBarPage,&s);
      s.clear();

      hScroll = (int) ((double)sourceWidth * hs);

      s.append(hScroll);
      updater.add(UpdateWindowList::actionHorizontalScrollBarPos,&s);
      s.clear();

    } else {

      if(hScrollBar == true) {

        s.append("off");
        updater.add(UpdateWindowList::actionHorizontalScrollBar,&s);
        s.clear();
        hScrollBar = false;
        hScroll = 0;

      }

    }

    if(sourceHeight > windowHeight) {

      if(vScrollBar == false) {

        s.append("on");
        updater.add(UpdateWindowList::actionVerticalScrollBar,&s);
        s.clear();
        vScrollBar = true;

      }

      s.append(sourceHeight);
      updater.add(UpdateWindowList::actionVerticalScrollBarRange,&s);
      s.clear();
      s.append(windowHeight);
      updater.add(UpdateWindowList::actionVerticalScrollBarPage,&s);
      s.clear();
      
      vScroll = (int) ((double)sourceHeight * vs);

      if(vScroll > sourceHeight - windowHeight)
        vScroll = sourceHeight - windowHeight;

      s.append(vScroll);
      updater.add(UpdateWindowList::actionVerticalScrollBarPos,&s);
      s.clear();


    } else {

      if(vScrollBar == true) {

        s.append("off");
        updater.add(UpdateWindowList::actionVerticalScrollBar,&s);
        s.clear();
        vScrollBar = false;
        vScroll = 0;
        s.append(vScroll);
        updater.add(UpdateWindowList::actionVerticalScrollBarPos,&s);
        s.clear();

      }

    }
    
  }

}

/**
 * This function is used by displayParagraphText to determine the size
 * in pixels of the rest of the word it is currently displaying. This is
 * used when in the middle of a word the style changes. For example:
 * \code
 * L<b>on</b>don
 * \endcode
 * So while displaying 'L' the size of 'ondon' must be known. That's 
 * what this function is for.
 * @param xcd  the chardata displayParagraphText is currently at
 * @param xbl  the xml code where the additional words could be
 * @param sl   the style list to be used
 * @return the length of the rest of the word in pixels
 */
int WMLWindow::additionalWordLength(XMLCharData *xcd, PointerList *xbl, StyleList *sl) {
  String s;

  xcd->savePosition();
  xcd->getNextWord(&s);
  xcd->recallPosition(); /* set the position of the char data back to before
                            the read word */
  
  if(s.equals(" ")) return 0;

  XMLBody *xb;
  bool lastCharacterWasSpace = false;
  int w = 0; /* the width that is going to be returned */
  
  while(true) {

    /* for every xmlbody in the body list calculate the width */
    /* when a space occurs, the algorithm stops */

    if(xbl->getSize() == 0) return w;

    xb = (XMLBody *)xbl->removeLastAdded();
    w += xmlLength(xb,sl,&lastCharacterWasSpace);

    if(lastCharacterWasSpace)
      return w;

  }

}

/**
 * This member calculates the width of a chain of XMLBody's, who are
 * chained together with their 'next' member. This function is used
 * by additionalWordLength.
 * @param xb  the starting point of the XMLBody chain
 * @param sl  the styles that must be used
 * @param lastCharacterWasSpace  when this boolean turns to true the 
 * algorithm can stop, it tells the calling function: 'the last 
 * character that i handled was a space so we can stop now'
 * @return the width
 */
int WMLWindow::xmlLength(XMLBody* xb, StyleList* sl, bool *lastCharacterWasSpace) {
  int w = 0; /* the width that is going to be returned */
  
  if(xb == NULL) {

    *lastCharacterWasSpace = false;
    sl->removeLastAdded();
    return 0;

  }

  if(xb->getContentType() == XMLBody::typeElement) {

    if(xb->getElement()->getTagName()->equals("do")       ||
       xb->getElement()->getTagName()->equals("fieldset") ||
       xb->getElement()->getTagName()->equals("select")   ||
       xb->getElement()->getTagName()->equals("input")    ||
       xb->getElement()->getTagName()->equals("img")      ||
       xb->getElement()->getTagName()->equals("table")    ||
       xb->getElement()->getTagName()->equals("go")       ||
       xb->getElement()->getTagName()->equals("prev")     ||
       xb->getElement()->getTagName()->equals("refresh")  ||
       xb->getElement()->getTagName()->equals("br") ) {

      *lastCharacterWasSpace = true;
      return 0;

    }

    if(xb->getElement()->getTagName()->equals("big"))
      sl->add(styleBig);

    if(xb->getElement()->getTagName()->equals("i"))
      sl->add(styleItalic);

    if(xb->getElement()->getTagName()->equals("b"))
      sl->add(styleBold);

    if(xb->getElement()->getTagName()->equals("small"))
      sl->add(styleSmall);

    if(xb->getElement()->getTagName()->equals("u"))
      sl->add(styleUnderlined);

    if(xb->getElement()->getTagName()->equals("em"))
      sl->add(styleEm);

    if(xb->getElement()->getTagName()->equals("strong"))
      sl->add(styleStrong);

    if(xb->getElement()->getTagName()->equals("anchor"))
      sl->add(styleLink);

    if(xb->getElement()->getTagName()->equals("a"))
      sl->add(styleLink);

    w = xmlLength(xb->getElement()->getBody(),sl,lastCharacterWasSpace);

  }

  if(xb->getContentType() == XMLBody::typeCharacterData) {

    xb->getCharacterData()->rewind();
    w = charDataLength(xb->getCharacterData(), sl->getStyle(),lastCharacterWasSpace);

  }

  if(*lastCharacterWasSpace)
    return w;

  return w + xmlLength(xb->getNextXMLBody(),sl,lastCharacterWasSpace);

}

/**
 * This member returns the width of the chardata in the argument from
 * the current position till the end of the chardata or till the first
 * space that it sees. When it sees a space, the lastCharacterWasSpace
 * variable is set to true.
 * @param xcd   the chardata
 * @param style  the style that will be used to display this chardata
 * @param lastCharacterWasSpace  this is set to true when the last 
 *        chracter that was handled by this function equaled a space
 */
int WMLWindow::charDataLength(XMLCharData *xcd, int style, bool *lastCharacterWasSpace) {

  String s;
  int w = 0;
  *lastCharacterWasSpace = true;

  while(true) {

    xcd->getNextWord(&s);

    if(s.equals(" "))
      return w; /* exit with lastCharacterWasSpace == true */

    if(s.equals("")) {

      *lastCharacterWasSpace = false;
      return w;

    }
    w += windowInterface->getStringWidth(&s,style);

  }
}

/**
 * Adds the given URI to the history. If adding the URI would cause 
 * the history to grow beyond 100 items, the first element is removed
 * from the history.
 * @param u the URI to add.
 */
void WMLWindow::historyAdd(URI *u) {
  String s;

  if(history->getSize() > 100) {
    
    /* Don't let the history grow beyond 100 items. */
    delete (URI *) history->getFirstAdded();
    history->removeFirstAdded();
    historyIndex--;
    
  }
  
  history->add(new URI(u->toString()));
  historyIndex++;
  
  /* Properly enable / disable back & forward buttons. */
  s.append((history->getSize() > 1 && historyIndex > 0) ? "on" : "off");
  updater.add(UpdateWindowList::actionBackButton, &s);
  
  s.clear();
  
  s.append(((signed int) history->getSize() > 1 && 
            historyIndex < (signed int) history->getSize() - 1) ? 
           "on" : "off");
  updater.add(UpdateWindowList::actionForwardButton, &s);
  
}
