/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WindowDrawingInterface.h,v 1.21 2001/02/28 14:10:11 wouter Exp $
 */

#ifndef __WINDOWDRAWINGINTERFACE_H__
#define __WINDOWDRAWINGINTERFACE_H__

class String;
class PointerList;
class MobilizerPreferences;
class GUIPeer;
class WMLGUIElement;
class WBMP;

/**
 * This class is an abstract class which WMLWindow uses to display
 * content on. It is abstract so the WMLWindow code can be used on
 * different devices. 
 * For example:
 * \code
 *   PalmInterface *pi = new PalmInterface();
 *   WMLWindow *ww = new WMLWindow((WindowDrawingInterface *) pi);
 *   ww->drawContentWindow();
 * \endcode
 */
class WindowDrawingInterface {

public:
/**
 * Tells the WindowDrawingInterface the location of the mobilizer 
 * preferences.
 *
 * @param mp the pointer to the MobilizerPreferences object
 */
  virtual void setPreferences(MobilizerPreferences *mp) = 0;

/** 
 * Tells the WindowDrawingInterface that WMLWindow is about to start
 * drawing. After this member has been called, calls to the other
 * members can be expected. When width and height are 0, actual drawing
 * won't take place, so WindowDrawingInterface::drawText won't be
 * called before WindowDrawingInterface::stopDrawing.
 *
 * @param width   the width of the content
 * @param height  the height of the content
 */
  virtual void startDrawing(int width, int height) = 0;

/** 
 * Tells the WindowDrawingInterface that WMLWindow is about to stop
 * drawing. After this member has been called, only
 * WindowDrawingInterface::startDrawing will be called.
 */
  virtual void stopDrawing() = 0;

/**
 * Returns the height of the string in the given WMLWindow::style in 
 * pixels.
 *
 * @param  s      the string
 * @param  style  the WMLWindow::style in which the string will be
 * displayed
 * @return the height in pixels
 */
  virtual int getStringHeight(String *s, int style) = 0;

/**
 * Returns the width of the string in the given WMLWindow::style in
 * pixels.
 *
 * @param  s      the string
 * @param  style  the WMLWindow::style in which the string will be
 * displayed
 * @return the width in pixels
 */
  virtual int getStringWidth(String *s, int style) = 0;

/** 
 * Draws the specified text in the display area assosiated with this
 * object.
 *
 * @param text    the to be displayed text
 * @param x       the x coordinate of the left-upper point at which to
 * display text
 * @param y       the y coordinate of the left-upper point at which to
 * display text
 * @param style   the WMLWindow::style that must be used to display text
 */
  virtual void drawText(String *text, int x, int y, int style) = 0;

/**
 * Returns the amount of pixels below the baseline of the font used
 * with the given WMLWindow::style.
 *
 * @param  style  the WMLWindow::style from which the pixel below 
 * baseline must be determined
 * @return amount of pixels below the baseline
 */
  virtual int getPixelsBelowBaseLine(int style) = 0;

/** 
 * Creates a new input element. This member must return a pointer to a
 * newly allocated platform dependent object that is subclassed from 
 * the GUIPeer class. The width and height members of this class must
 * be set to the correct values, meaning: the width and height of the
 * input element.
 * @return the platform dependent object
 */
  virtual GUIPeer *createInputElement() = 0;

/** 
 * Draws an input element in the display area assosiated with this 
 * object. The WMLGUIElement in the argument has members like 
 * WMLGUIElement::getHeight() and WMLGUIElement::getXCoordinate().
 *
 * @param wie the WMLGUIElement
 */
  virtual void drawInputElement(WMLGUIElement *wie) = 0;

/** 
 * Creates a new select element. This member must return a pointer to
 * a newly allocated platform dependent object that is subclassed from
 * the GUIPeer class. The width and height members of this class must
 * be set to the correct values, meaning: the width and height of the
 * select element.
 *
 * @param  optionList a pointerlist of String object containing the
 * options, the objects in the pointerlist (the Strings) are expected
 * to be deleted in this function
 * @return the platform dependent object
 */
  virtual GUIPeer *createSelectElement(PointerList *optionList) = 0;

/** 
 * Creates a new image peer. This member must return a pointer to
 * a newly allocated platform dependent object that is subclassed from
 * the GUIPeer class. The width and height members of this class must
 * be set to the correct values, meaning: the width and height of the
 * image.
 *
 * @param  wbmp in this obect the data of the wbmp is located
 * @return the platform dependent object
 */
  virtual GUIPeer *createImage(WBMP *wbmp) = 0;

/** 
 * Draws a select element in the display area assosiated with this 
 * object. The WMLGUIElement in the argument has members like
 * WMLGUIElement::getHeight() and WMLGUIElement::getXCoordinate().
 *
 * @param wse the WMLGUIElement
 */
  virtual void drawSelectElement(WMLGUIElement *wse) = 0;

/** 
 * Draws the specified image in the display area assosiated with this
 * object.
 *
 * @param image   the image to display
 */
  virtual void drawImage(WMLGUIElement *image) = 0;

};

#endif
