/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: CharStream.cpp,v 1.12 2001/01/27 14:36:40 wouter Exp $ 
 *
 * Reviewed  3-12-2000 ph (one-line if's)
 * Reviewed 16-01-2001 wk
 */

#include <string.h>

#include "CharStream.h"
#include "../lib/String.h"
#include "XMLEndOfFileException.h"

/** 
 * Constructs a new CharStream object from the String argument.
 * @param source the source from which the character stream must be
 * built
 */
CharStream::CharStream(String *source) {

  /* copy the source: */
  begin = new char[source->getLength()+1];
  strcpy(begin, source->getPtr());

  current = begin;
  last = begin + source->getLength() - 1;

  currentLineNumber = 1;
  recall = begin;

}

/**
 * Delete the object
 */
CharStream::~CharStream() {

  delete begin;

}

/**
 * This function tests the code of this class
 *
 * @return true when all tests succeed, otherwise false
 */
bool CharStream::test() {

  /* testing getNextChar() */
  String s("a");
  CharStream cs(&s);

  if(cs.getNextChar() != 'a')
    return false;

  bool exceptionThrown = false;

  try {

    cs.getNextChar();

  } catch (XMLEndOfFileException) {
    
    exceptionThrown = true;

  }

  if(!exceptionThrown)
    return false;

  /* testing pushBack()/getLineNumber() */
  s.clear();
  s.append("a\nb");
  CharStream cs2(&s);

  cs2.getNextChar();
  if(cs2.getLineNumber() != 1)
    return false;

  cs2.getNextChar();
  if(cs2.getLineNumber() != 2)
    return false;

  cs2.pushBack();
  if(cs2.getLineNumber() != 1)
    return false;

  cs2.pushBack();

  if(cs2.getNextChar() != 'a')
    return false;

  /* testing isEndOfFile() */
  s.clear();
  s.append("a\n\r\t    ");
  CharStream cs3(&s);

  if(cs3.isEndOfFile() == true)
    return false;

  cs3.getNextChar();

  if(cs3.isEndOfFile() == false)
    return false;

  /* testing savePosition()/recallPosition() */
  s.clear();
  s.append("1234");
  CharStream cs4(&s);

  cs4.savePosition();

  cs4.getNextChar();
  cs4.getNextChar();
  cs4.recallPosition();

  if(cs4.getNextChar() != '1')
    return false;

  return true;

}

/**
 * Returns the next character from the character stream.
 *
 * @return the character
 */
char CharStream::getNextChar() {

  if(current > last) 
    throw XMLEndOfFileException();

  char character = current[0];
  current++;

  if(character == '\n') 
    currentLineNumber++;

  return character;

}

/**
 * Pushes the last character from the stream back on the stream.
 */
void CharStream::pushBack() {

  current--;

  if(current < begin) 
    current = begin;

  if(current[0] == '\n') 
    currentLineNumber--;
  
}

/**
 * Returns the line number of the line the stream is currently on.
 *
 * @return the line number
 */
int CharStream::getLineNumber() {

  return currentLineNumber;

}

/** 
 * Tests if the character stream is on the end of the file
 * (after seeing spaces, tabs, newlines, ...).
 *
 * @return true if the character stream is on the end of the file, 
 * otherwise false
 */
bool CharStream::isEndOfFile() {
  char *walker = current;

  while(walker <= last) {

    if(*walker != ' ' && *walker != '\n' && *walker != '\r' && *walker != '\t')
      return false;    /* when the character is not one of the above 
			                    the stream is not at the end */
    
    walker++;

  }
  return true;

}

/**
 * Saves the current position of the tokenizer. When the
 * recallPosition member is called the tokenizer jumps back to this
 * position.  
 */
void CharStream::savePosition() {

  recall = current;

}

/**
 * The tokenizer jumps back to the position it was in when
 * savePosition was called.
 * 
 */
void CharStream::recallPosition() {

  current = recall;

}
