/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: XMLBody.cpp,v 1.8 2001/01/27 14:36:40 wouter Exp $ 
 *
 * Reviewed 4-12-2000 ph (enum, NULLs)
 */

#include <stdlib.h>

#include "../lib/String.h"
#include "XMLBody.h"
#include "XMLCharData.h"
#include "XMLElement.h"
#include "XMLParser.h"
#include "XMLTokenizer.h"

/**
 * Constructs a new XMLBody. 
 */
XMLBody::XMLBody() {

  element = NULL;
  characterData = NULL;
  contentType = typeNothing;
  nextElement = NULL;

}

/** 
 * Constructs a new XMLBody with the same elements as the XMLBody in
 * the argument.  
 * @param the original XMLBody
 */
XMLBody::XMLBody(XMLBody *xb) {

  if(xb->getElement() != NULL)
    element = new XMLElement(xb->getElement());
  else
    element = NULL;

  if(xb->getCharacterData() != NULL)
    characterData = new XMLCharData(xb->getCharacterDataAsString());
  else
    characterData = NULL;

  if(element != NULL) 
    contentType = typeElement;

  if(characterData != NULL) {

    contentType = typeCharacterData;
    characterData->setStartLineNumber(xb->getCharacterData()->getStartLineNumber());
    characterData->setEndLineNumber(xb->getCharacterData()->getEndLineNumber());

  }

  if(element == NULL && characterData == NULL) 
    contentType = typeNothing;

  if(xb->getNextXMLBody() != NULL) 
    nextElement = new XMLBody(xb->getNextXMLBody());
  else
    nextElement = NULL;

}

/**
 * Delete the object
 */
XMLBody::~XMLBody() {

  delete element;
  delete nextElement;
  delete characterData;

}

/**
 * This function tests the code of this class
 *
 * @return true when all tests succeed, otherwise false
 */
bool XMLBody::test() {
  String *s = new String("<tag>abc<link>def</link>ghi</tag>");
  XMLElement *xe = new XMLElement();
  XMLParser *xp = new XMLParser(s, xe);
  XMLBody *xb = xe->getBody();

  delete s;

  if(xb->getContentType() != typeCharacterData) return false;

  s = xb->getCharacterDataAsString();

  if(!s->equals("abc")) return false; 

  xb = xb->getNextXMLBody();

  if(xb->getContentType() != typeElement) return false;

  xb = xb->getNextXMLBody();

  if(xb->getContentType() != typeCharacterData) return false;

  s = xb->getCharacterDataAsString();

  if(!s->equals("ghi")) return false;
  
  delete xe;

  delete xp;

  return true;

}

/** 
 * Appends the body in the argument to the body
 *
 * @param xb the XMLBODY *element to be appended 
 */
void XMLBody::append(XMLBody *xb) {

  if(xb->getContentType() != typeNothing) { 
    
    if(nextElement == 0) {
      
      nextElement = new XMLBody(xb);
      
    } else {
      
      nextElement->append(xb);
      
    }

  }

}

/** 
 * This function is part of the xml parser. It reads the
 * <pre>
 *  hallo!<br/> </link>
 * </pre>
 * part of
 * <pre>
 *  <link show = 'on'> hallo!<br/> </link>
 * </pre>
 * @param tag the tagname of the body the parser is in
 * @param tokenizer the tokenizer to use
 * @param parser the parser, this parameter is used to add warnings
 * and errors
 * 
 */
void XMLBody::parse(String *tag, XMLTokenizer *tokenizer, XMLParser *parser) {

  tokenizer->getNextToken();

  parser->readCommentsAndProcessingInstructions();

  if(tokenizer->getLastReadToken() == XMLTokenizer::tokenTagClose) {

    if(tokenizer->getTokenAsString()->equals(tag))
      return;

    /* the wrong closing tag has been read */
    parser->appendToErrorString("line ");
    parser->appendToErrorString(tokenizer->getLineNumber());
    parser->appendToErrorString(": xml error: the '<");
    parser->appendToErrorString(tag->getPtr());
    parser->appendToErrorString(">' tag must be closed before '</");
    parser->appendToErrorString(tokenizer->getTokenAsString()->getPtr());
    parser->appendToErrorString(">'.");
    parser->generateError();
    return;

  }

  if(tokenizer->getLastReadToken() == XMLTokenizer::tokenTagOpen) {

    /* a new open tag has been read so we fill this new element */
    contentType = typeElement;
    element = new XMLElement();
    element->parse(tokenizer->getTokenAsString(), tokenizer, parser);

  } else {

    /* there was no open tag so this must be character data */
    contentType = typeCharacterData;
    String *temp = new String();

    int i = tokenizer->getLineNumber();

    if(tokenizer->getLastReadToken() == XMLTokenizer::tokenCDataOpen) {

      tokenizer->readCDataSection(temp);
      characterData = new XMLCharData(temp);
      characterData->setMustReplaceEntities(false);

    } else {

      /* first we push back the last token because it is part of the
         character data */

      tokenizer->pushbackToken();
      if(tokenizer->readCharacterData(temp))
        parser->generateUnexpectedTokenError();

      characterData = new XMLCharData(temp);
      characterData->setMustReplaceEntities(true);

    }

    characterData->setStartLineNumber(i);
    characterData->setEndLineNumber(tokenizer->getLineNumber());

    delete temp;

  }
  
  nextElement = new XMLBody();
  /* fill the next xml body: */
  nextElement->parse(tag, tokenizer, parser);

  if(nextElement->contentType == typeNothing) {

    /* it there was nothing to fill the next xmlbody is empty, so delete it */
    delete nextElement;
    nextElement = NULL;

  }

}

/**
 * Gets the content type of this element in the element body.
 *
 * @return XMLBody::element, XMLBody::characterData
 */
XMLBody::contentTypes XMLBody::getContentType() {

  return contentType;

}

/** 
 * Gets the element of this body element when the element is a
 * XMLElement.  Returns NULL when element is character data.
 *
 * @return returns a pointer to the XMLElement, do not delete 
 */
XMLElement *XMLBody::getElement() {

  return element;

}

/** 
 * Gets the body element of which the element name equals the name in
 * the argument.  Returns NULL when the "name" tag is not in the body.
 *
 * @return returns a pointer to the XMLElement, do not delete 
 */
XMLBody *XMLBody::getElement(char *name) {

  if(contentType == typeCharacterData) {

    if(nextElement == NULL)
      return NULL;
    return nextElement->getElement(name);

  }

  if(element == NULL) 
    return NULL;

  if(element->getTagName()->equals(name))
    return this;

  if(nextElement == NULL)
    return NULL;

  return nextElement->getElement(name);

}

/** 
 * Gets the character data of this body element when the element is
 * character data.  Returns NULL when the element is an XMLElement.
 *
 * @return returns a pointer to a String object containing the
 * charater data, do not delete
 */
String *XMLBody::getCharacterDataAsString() {
  
  if (characterData == NULL) return NULL;
  return characterData->getCharDataAsString();

}

/** 
 * Gets the character data of this body element when the element is
 * character data.  Returns NULL when the element is an XMLElement.
 *
 * @return returns a pointer to a XMLCharData object containing the
 * charater data, do not delete 
 */
XMLCharData *XMLBody::getCharacterData() {

  return characterData;

}

/**
 * Gets the the next element of body.
 *
 * @return returns a pointer to a XMLBody, do not delete
 */
XMLBody *XMLBody::getNextXMLBody() {

  return nextElement;

}

/**
 * The string in the argument is filled with the parse tree in text.
 *
 * @param s the string that will be filled with the parse tree
 */
void XMLBody::getParseTreeAsString(int depth, String *s) {

  if(contentType == typeElement)
    element->getParseTreeAsString(depth, s);

  if(contentType == typeCharacterData) {

    int i;
    for(i = depth; i > 0; i--)
      s->append(' ');

    s->append('\"');
    s->append(characterData->getCharDataAsString());
    s->append("\"\n");

  }
  
  if(nextElement != 0)
    nextElement->getParseTreeAsString(depth, s);

}
