/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: XMLCharData.cpp,v 1.9 2001/01/27 14:36:40 wouter Exp $ 
 *
 * Reviewed 4-12-2000 ph 
 */

#include <stdlib.h>

#include "../lib/String.h"
#include "XMLCharData.h"

/** 
 * Constructs a new XMLCharData. 
 * @param s the string containing the character data
 */
XMLCharData::XMLCharData(String *s) {

  characterData = new String(s);
  replaceEntities = true;
  startLineNumber = 0;
  endLineNumber = 0;
  currentPosition = 0;
  savedPosition = 0;

}

/**
 * Delete the object
 */
XMLCharData::~XMLCharData() {

  delete characterData;

}

/** 
 * This function tests the code of this class
 *
 * @return true when all tests succeed, otherwise false
 */
bool XMLCharData::test() {

  String s(" abc def ");
  XMLCharData xcd(&s);

  s.clear();

  xcd.getNextWord(&s);

  if(!s.equals(" ")) 
    return false;

  xcd.getNextWord(&s);

  if(!s.equals("abc")) 
    return false;

  xcd.getNextWord(&s);

  if(!s.equals(" ")) 
    return false;

  xcd.getNextWord(&s);

  if(!s.equals("def")) 
    return false;

  xcd.getNextWord(&s);

  if(!s.equals(" ")) 
    return false;

  xcd.getNextWord(&s);

  if(s.getLength() != 0) 
    return false;

  return true;

}

/** 
 * Returns the content of the character data
 *
 * @return the character data as a string, do not delete!
 */
String *XMLCharData::getCharDataAsString() {

  return characterData;

}

/** 
 * Sets the starting line of the character data
 *
 * @param i the line number
 */
void XMLCharData::setStartLineNumber(int i) {

  startLineNumber = i;

}

/** 
 * Returns the starting line of the character data
 *
 * @return the line number
 */
int XMLCharData::getStartLineNumber() {

  return startLineNumber;

}

/** 
 * Sets the ending line of the character data
 *
 * @param i the line number
 */
void XMLCharData::setEndLineNumber(int i) {

  endLineNumber = i;

}

/** 
 * Returns the starting line of the character data
 *
 * @return the line number
 */
int XMLCharData::getEndLineNumber() {

  return endLineNumber;

}

/** 
 * Determines whether the chardata must be checked on entities
 *
 * @return true if the entities must be replaced
 */
bool XMLCharData::getMustReplaceEntities() {

  return replaceEntities;

}

/** 
 * Sets this value which determines whether this character data must
 * be checked for entities
 *
 * @param b true if the entities must be replaced 
 */
void XMLCharData::setMustReplaceEntities(bool b) {

  replaceEntities = b;

}

/**
 * Returns the next word from the character data, when s is empty,
 * there are no words left
 *
 * @param s the string thar will be filled with the next word 
 */
void XMLCharData::getNextWord(String *s) {

  int length = characterData->getLength();
  char *p, *begin = characterData->getPtr(); 

  s->clear();

  if(currentPosition < length) {
    p = begin + currentPosition;

    /* other whitespace (/r/t, etc.) have been filterd out by the tokenizer */
    if(*p == ' ') {

      while(*p == ' ' && p - begin < length)
        p++;
    
      s->append(' ');

    } else {

      while(*p != ' ' && p - begin < length) {

        s->append(*p);
        p++;

      }

    }

    currentPosition = p - begin;
  }

}

/**
 * Saves the position of the pointer in the character data. When
 * recallPosition is called the position is changed to the position
 * the pointer was on when sacePosition was called.  
 */
void XMLCharData::savePosition() {

  savedPosition = currentPosition;

}

/**
 * Saves the position of the pointer in the character data. When recallPosition is called
 * the position is changed to the position the pointer was on when sacePosition was called.
 */
void XMLCharData::recallPosition() {

  currentPosition = savedPosition;

}

/**
 * Sets the position of the pointer in the character data to 0, so getNextWord will return the
 * first word. 
 */
void XMLCharData::rewind() {

  currentPosition = 0;

}

