/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: XMLParser.cpp,v 1.13 2001/03/09 12:42:19 wouter Exp $ 
 *
 * Reviewed 4-12-2000 ph (catch blocks, exit points, enum)
 */

#include <stdlib.h>
#include <stdio.h>

#include "../lib/String.h"
#include "XMLElement.h"
#include "XMLEndOfFileException.h"
#include "XMLParser.h"
#include "XMLParseErrorException.h"

/** 
 * Constructs a new XMLParser.
 *
 * @param source the source which must be parsed
 */
XMLParser::XMLParser(String *source, XMLElement *topElement) {

  xmlTokenizer = new XMLTokenizer(source);
  xmlErrorMessage = new String();
  xmlTopElementName = NULL;

  try {

    xmlTokenizer->getNextToken();

    readXMLDeclaration();
    readCommentsAndProcessingInstructions();
    readDocTypeDeclaration();
    readCommentsAndProcessingInstructions();

    if(xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenTagOpen)
      generateUnexpectedTokenError();

    topElement->parse(xmlTokenizer->getTokenAsString(), xmlTokenizer, this);

    if(xmlTopElementName != 0)
      if(!xmlTopElementName->equals(topElement->getTagName())) {

        appendToErrorString("xml error: the root tag declared in the Document Dype Declaration (<");
        appendToErrorString(xmlTopElementName->getPtr());
        appendToErrorString(">) is not the same as the real root tag (<");
        appendToErrorString(topElement->getTagName()->getPtr());
        appendToErrorString(">).");
        generateError();

      }

  } catch (XMLParseErrorException) {

    return;

  } catch (XMLEndOfFileException) {

    xmlErrorMessage->append("xml error: unexpected end of file. started reading at line ");
    xmlErrorMessage->append(xmlTokenizer->getTokenStartLineNumber());
    return;

  }

  try {

    xmlTokenizer->getNextToken();

  } catch (XMLEndOfFileException) {

    if(xmlTokenizer->getTokenAsString()->getLength() == 0) 
      return;

  }

  try {

    readCommentsAndProcessingInstructions();

  } catch (XMLEndOfFileException) {

    if(xmlTokenizer->getTokenAsString()->getLength() == 0) 
      return;

  }

  xmlErrorMessage->append("line ");
  xmlErrorMessage->append(xmlTokenizer->getLineNumber());
  xmlErrorMessage->append(": xml error: character data after last closing tag.");

}

/**
 * Delete the object
 */
XMLParser::~XMLParser() {

  delete xmlTokenizer;
  delete xmlErrorMessage;
  delete xmlTopElementName;

}

/** 
 * This function tests the code of this class
 *
 * @return true when all tests succeed, otherwise false
 */
bool XMLParser::test() {
  String *s = new String("<?xml version=\"1.0\"?>\r\n<!DOCTYPE wml PUBLIC \"-//WAPFORUM//DTD WML 1.2//EN\"\r \"http://www.wapforum.org/DTD/wml12.dtd\">\r\n<wml/>");

  XMLElement *xe = new XMLElement();
  XMLParser *xp = new XMLParser(s, xe);
  delete s;

  if(xp->getTopElementName()->equals(xe->getTagName()) != true) return false; /* leaks xp, xe */

  delete xp;
  delete xe;
  return true;

}

/**
 * Gives access to the string containing the name of the top element
 * according to the Document Type Declaration.  
 */
String *XMLParser::getTopElementName() {

  return xmlTopElementName;

}

/**
 * Returns a string containing the error message of the parser. When
 * the parser did not generate an error this function returns NULL.
 *
 * @return a string containing the error message or NULL, do not delete 
 */
String *XMLParser::getErrorMessage() {
  
   /* this function must return null when 
      there is no error */
  if  (xmlErrorMessage->getLength() == 0) 
    return NULL;

  return xmlErrorMessage;

}

/** 
 * Appends to the error string.
 *
 * @param error the error text
 */
void XMLParser::appendToErrorString(char *error) {

  xmlErrorMessage->append(error);

}

/**
 * Appends to the error string
 *
 * @param i the integer to append
 */
void XMLParser::appendToErrorString(int i) {

  xmlErrorMessage->append(i);

}

/** 
 * Generate an error. 
 *
 * @param error the error text
 */
void XMLParser::generateError(char *error) {

  makeErrorText (xmlErrorMessage, error, XMLTokenizer::tokenNothing);
  throw XMLParseErrorException();

}

/** 
 * Generate an error, without adding to error text.
 */
void XMLParser::generateError() {

  throw XMLParseErrorException();

}

/** 
 * Generate an error and mention the last read token at the end of
 * the error text.
 *
 * @param error the error text
 */
void XMLParser::generateErrorMentionToken(char *error) {

  makeErrorText (xmlErrorMessage, error, xmlTokenizer->getLastReadToken());
  throw XMLParseErrorException();

}

/** 
 * Generate an error, with text "unexpected token: " and then the last
 * read token.
 */
void XMLParser::generateUnexpectedTokenError() {

  generateErrorMentionToken("unexpected token: ");

}

/** 
 * Reads the <!-- comment --!> and <?pi eqwe?> parts
 */
void XMLParser::readCommentsAndProcessingInstructions() {

  while(xmlTokenizer->getLastReadToken() == XMLTokenizer::tokenComment || 
        xmlTokenizer->getLastReadToken() == XMLTokenizer::tokenProcessingInstruction)
    xmlTokenizer->getNextToken();

}

/**
 * Reads the <?xml version="1.0"?> part.
 */
void XMLParser::readXMLDeclaration() {

  /* i can't really get rid of all the exit points in this function -wk */

  if(xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenXMLOpen)
    return; /* exit point */

  xmlTokenizer->getNextToken();  
  
  if(xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenName)
    generateError("after '<?xml' should follow 'version=...'.");
    /* exit point */

  String *s = xmlTokenizer->getTokenAsString();

  if(!s->equals("version"))
    generateError("after '<?xml' should follow 'version=...'."); /* exit point */

  while(xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenXMLClose && 
	      xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenCloseTag) {

    /* read all attributes of <?xml> tag */
    if(xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenName)  
      /* the attribute name */
      generateUnexpectedTokenError(); 

    if(!s->equals("standalone") && 
       !s->equals("encoding") && 
       !s->equals("version"))
      generateErrorMentionToken("unexpected attribute in <?xml>: "); /* exit point */
    
    xmlTokenizer->getNextToken();

    if(xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenEquals) 
      /* the '=' */
      generateUnexpectedTokenError(); /* exit point */
    
    xmlTokenizer->getNextToken();
    if(xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenStringConstant)
      /* the string */
      generateUnexpectedTokenError(); /* exit point */

    xmlTokenizer->getNextToken();

  }
  
  if(xmlTokenizer->getLastReadToken() == XMLTokenizer::tokenCloseTag)
    generateError("XML declaration should end with '?>' in stead of '>'.");
    /* exit point */

  xmlTokenizer->getNextToken();

  /* exit point */

}

/**
 * Reads the <!DOCTYPE root "bla" "bla"> part 
 */
void XMLParser::readDocTypeDeclaration() {

  if(xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenDocTypeOpen) 
    return;

  if(xmlTokenizer->getNextToken() != XMLTokenizer::tokenName)
    generateUnexpectedTokenError();
  
  xmlTopElementName = new String(xmlTokenizer->getTokenAsString());
  
  if(xmlTokenizer->getNextToken() != XMLTokenizer::tokenName)
    generateUnexpectedTokenError();

  if(!xmlTokenizer->getTokenAsString()->equals("PUBLIC"))
    generateError("this parser is only built for PUBLIC dtd's.");

  xmlTokenizer->getNextToken();

  if(xmlTokenizer->getLastReadToken() != XMLTokenizer::tokenStringConstant)
    generateUnexpectedTokenError();

  if(xmlTokenizer->getNextToken() != XMLTokenizer::tokenStringConstant)
    generateUnexpectedTokenError();

  if(xmlTokenizer->getNextToken() != XMLTokenizer::tokenCloseTag)
    generateUnexpectedTokenError();
  
  xmlTokenizer->getNextToken();
}

/**
 * Generates the error text
 * @param s the string to append to
 * @param message the error text
 * @param token the token that was last read
 */
void XMLParser::makeErrorText(String *s, char *message, XMLTokenizer::xmlToken token) {

  s->append("line ");
  s->append(xmlTokenizer->getLineNumber());
  s->append(": xml error: ");
  s->append((char *) message);
  
  if(token != XMLTokenizer::tokenNothing) {

    switch (token) {

    case XMLTokenizer::tokenStringConstant:
      s->append("'\"");
      s->append(xmlTokenizer->getTokenAsString()->getPtr());
      s->append("\"'.");
      break;

    case XMLTokenizer::tokenTagOpen:
      s->append("'<'.");
      break;

    case XMLTokenizer::tokenTagClose:
      s->append("'</'.");
      break;

    case XMLTokenizer::tokenComment:
      s->append("comment.");
      break;

    case XMLTokenizer::tokenProcessingInstruction:
      s->append("a processing instruction.");
      break;

    case XMLTokenizer::tokenEndlessString:
      s->append("an endless string.");
      break;

    default:
      s->append("'");
      s->append(xmlTokenizer->getTokenAsString()->getPtr());
      s->append("'.");
      break;

    }

  }

}
