/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: VarPool.cpp,v 1.4 2001/01/27 14:36:41 wouter Exp $
 */

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "VarPool.h"

/**
 * CreateVarPool() creates a variable pool and returns a pointer to it.
 *
 * @return pointer to VarPool or NULL on error.
 */
VarPool* CreateVarPool() {
  VarPool* vp;

  vp = (VarPool*) malloc(sizeof(VarPool));

  if(vp) {

    vp->var = NULL;
    vp->value = NULL;
    vp->next = NULL;
  
  }

  return vp;

}

/**
 * DestroyVarPool() frees all memory used by the variable pool.
 * It is the reponsibility of the caller to release any memory
 * in use by variable values.
 *
 * @param vp pointer to VarPool.
 * @return 0.
 */
int DestroyVarPool(VarPool* vp) {
  VarPool* vp_prev;

  while(vp) {

    if(vp->var)
      free(vp->var);
    vp_prev = vp;
    vp = vp->next;
    free(vp_prev);

  }

  return 0;
}

/**
 * CreateVar() creates a new variable in the variable pool. The
 * supplied variable name is copied and remains the responsibility
 * of the caller.
 *
 * @param vp pointer to VarPool.
 * @param var pointer to null-terminated variable name.
 * @return 0 on success, EVAREXISTS if the variable already exists,
 * ENULLSYM if vp is NULL, EOUTOFMEMORY if memory could not be
 * allocated, ENULLVAR if var is NULL.
 */
int CreateVar(VarPool* vp, char *var) {
  VarPool *vp_prev;

  if(!vp) return ENULLSYM;
  if(!var) return ENULLVAR;

  do {

    if(!vp->var) {

      vp->var = (char *) malloc(strlen(var) + 1);
      if(!vp->var) 
	return EOUTOFMEMORY;
      strcpy(vp->var, var);
      return 0;

    }

    if(strcmp(var, vp->var) == 0) return EVAREXISTS;
    
    vp_prev = vp;
    vp = vp->next;
    
  } while(vp);
  
  vp = (VarPool *) malloc(sizeof(VarPool));
  if(!vp) return EOUTOFMEMORY;
  vp->var = (char *) malloc(strlen(var) + 1);
  if(!vp->var) return EOUTOFMEMORY;

  strcpy(vp->var, var);
  vp->value = NULL;
  vp->next = NULL;
  vp_prev->next = vp;
  
  return 0;

}

/**
 * GetVarValue() returns the value of a variable.
 *
 * @param var variable whose value is to be retrieved.
 * @return pointer to variable value, NULL if variable was not
 * found or was uninitialized.
 */
void* GetVarValue(VarPool* vp, char* var) {
  
  if(!vp) return NULL;
  if(!var) return NULL;

  do {

    if(vp->var && strcmp(var, vp->var) == 0) return vp->value;
    vp = vp->next;

  } while(vp);

  return NULL;

}

/**
 * ExistsVar() tests whether a variable exists in the variable pool.
 *
 * @param vp pointer to variable pool.
 * @param var pointer at null-terminated variable name.
 * @return 0 if variable exists in vp, ENULLSYM if vp is NULL, 
 * ENULLVAR if var is NULL, EVARNOTFOUND if variable could not 
 * be found in vp. 
 *
 * Bug: ExistsVar() does not distinguish between nonexisting
 * and uninitialized variables; in both cases it returns 
 * EVARNOTFOUND. The remedy is to always initialize your 
 * variables before calling ExistsVar().
 */
int ExistsVar(VarPool* vp, char* var) {
  
  if(!vp) return ENULLSYM;
  if(!var) return ENULLVAR;

  return(GetVarValue(vp, var) == NULL ? EVARNOTFOUND : 0);

}

/**
 * SetVarValue() assigns a value to an already existing variable
 * in the variable pool. Any previous value is overwritten; cleaning
 * up the previous value is the responsibility of the caller. The
 * supplied value also remains the responsibility of the caller.
 *
 * @param vp pointer to variable pool.
 * @param var pointer to null-terminated variable name.
 * @param value pointer to value.
 * @return 0 on success, ENULLSYM if vp is NULL, ENULLVAR if 
 * var is NULL, ENULLVALUE if value is NULL, or EVARNOTFOUND if
 * var does not exist in vp.
 */
int SetVarValue(VarPool* vp, char* var, void* value) {
  VarPool* vp_prev;
  
  if(!vp) return ENULLSYM;
  if(!var) return ENULLVAR;
  if(!value) return ENULLVALUE;

  do {

    if(vp->var && strcmp(vp->var, var) == 0) {

      vp->value = value;
      return 0;

    }

    vp_prev = vp;
   
    vp = vp->next;

  } while(vp);
  
  return EVARNOTFOUND;

}

