/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptFunctionPool.cpp,v 1.5 2001/01/27 14:36:41 wouter Exp $
 */

#include <stdio.h>
#include <assert.h>

#include "VarPool.h"
#include "WMLScriptVarPool.h"
#include "WMLScriptValueObject.h"
#include "WMLScriptParseTreeNodes.h"
#include "WMLScriptFunction.h"
#include "WMLScriptPragma.h"
#include "WMLScriptParseResult.h"
#include "WMLScriptParseTree.h"
#include "WMLScriptFunctionPool.h"

#undef WMLS_DEBUG

/**
 * A WMLScriptFunctionPool encapsulates the function table returned 
 * by the parser. A function table consists of three items:
 * 
 * 1. The parse tree for a function
 * 2. The parameter declaration for a function
 * 3. The special flags for a function
 * 
 * You should never need to instantiate this class directly -- it is
 * done by WMLScriptParseContext.
 * @param result the parse results as returned by the parser.
 */
WMLScriptFunctionPool::WMLScriptFunctionPool(const WMLScriptParseResult *parseResult) {
  int i;
  numberOfFunctions = parseResult->numberOfFunctions;
  functionTable = parseResult->functionTable;

  parseTrees = 
    (WMLScriptParseTree **) malloc(sizeof(WMLScriptParseTree *) * 
				   numberOfFunctions);
  for(i = 0; i < numberOfFunctions; i++)
    parseTrees[i] = 
      new WMLScriptParseTree(functionTable[i].parseTree);

  varPools = 
    (WMLScriptVarPool **) malloc(sizeof(WMLScriptVarPool *) * 
				 numberOfFunctions);

  for(i = 0; i < numberOfFunctions; i++)
    varPools[i] = 
      new WMLScriptVarPool(functionTable[i].varPool);

  flags = (char *) malloc(sizeof(char) * numberOfFunctions);
  for(i = 0; i < numberOfFunctions; i++)
    flags[i] = functionTable[i].flags;
}

/**
 * Destroys the object. You should never need to call this
 * destructor directly -- it is done by WMLScriptParseContext.
 */
WMLScriptFunctionPool::~WMLScriptFunctionPool() {
  int i;

  if(flags)
    free(flags);

  if(varPools) {
    for(i = 0; i < numberOfFunctions; i++)
      delete varPools[i];
    free(varPools);
  }
  
  if(parseTrees) {
    for(i = 0; i < numberOfFunctions; i++)
      delete parseTrees[i];
    free(parseTrees);
  }

  assert(functionTable != NULL);

  for(i = 0; i < numberOfFunctions; i++)
    if(functionTable[i].name)
      free(functionTable[i].name);
  
  free(functionTable);

}

/**
 * Returns the parse tree for a given function by name.
 * @param s the name of the function.
 * @return pointer at the requested WMLScriptParseTree object
 * or NULL.
 */
WMLScriptParseTree *WMLScriptFunctionPool::getParseTree(char *s) const {
  int i;
  for(i = 0; i<numberOfFunctions; i++) 
    if(strcmp(s, functionTable[i].name->id.name) == 0) 
      return parseTrees[i];

  return NULL;
}


/**
 * Returns the variable pool (parameter list) for a given function 
 * by name.
 * @param s the name of the function.
 * @return pointer at the requested WMLScriptVarPool object
 * or NULL.
 */
WMLScriptVarPool *WMLScriptFunctionPool::getVarPool(char *s) const {
  int i;
  for(i = 0; i<numberOfFunctions; i++) 
    if(strcmp(s, functionTable[i].name->id.name) == 0) 
      return varPools[i];

  return NULL;
}

/**
 * Returns the flags for a given function by name.
 * @param s the name of the function.
 * @return function flags (see WMLScriptFunction.h) or zero (0).
 */
char WMLScriptFunctionPool::getFlags(char *s) const {
  int i;
  for(i = 0; i<numberOfFunctions; i++)
    if(strcmp(s, functionTable[i].name->id.name) == 0)
      return flags[i];

  return 0;
}

/**
 * Returns the number of functions in this pool.
 * @return number of functions.
 */
int WMLScriptFunctionPool::size() const {
  return numberOfFunctions;
}

/** 
 * Returns the name of a given function by index.
 * @param i index.
 * @return pointer at C string.
 */
char *WMLScriptFunctionPool::getName(int i) const {
  return functionTable[i].name->id.name;
}

void WMLScriptFunctionPool::dumpFunctions() const {
  int i;
  printf("WMLScriptFunctionPool %p dumping functions\n", this);
  printf("number of functions: %d\n", numberOfFunctions);
  for(i = 0; i < numberOfFunctions; i++) {
    printf("function: %s\n", functionTable[i].name->id.name);
  }
  printf("WMLScriptFunctionPool dumped functions\n");
}

