/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptPragmaPool.cpp,v 1.5 2001/01/27 14:36:42 wouter Exp $
 */

#include <stdio.h>
#include <assert.h>

#include "WMLScriptValueObject.h"
#include "VarPool.h"
#include "WMLScriptVarPool.h"
#include "WMLScriptParseTreeNodes.h"
#include "WMLScriptFunction.h"
#include "WMLScriptPragma.h"
#include "WMLScriptParseResult.h"
#include "WMLScriptParseTree.h"
#include "WMLScriptPragmaPool.h"

#undef WMLS_DEBUG

/**
 * A WMLScriptPragmaPool encapsulates the pragmas returned 
 * by the parser.
 * @param result the parse results as returned by the parser.
 */
WMLScriptPragmaPool::WMLScriptPragmaPool(const WMLScriptParseResult *parseResult) {
  pragmaTable = parseResult->pragmaTable;
  numberOfPragmas = parseResult->numberOfPragmas;
}

/**
 * Destroys the object. You should never need to call this
 * destructor directly -- it is done by WMLScriptParseContext.
 */
WMLScriptPragmaPool::~WMLScriptPragmaPool() {
  int i;

  for(i = 0; i < numberOfPragmas; i++) {

    switch(pragmaTable[i]->type) {

    case typeExternalCompilationUnit:
      assert(pragmaTable[i]->ext.id != NULL);
      assert(pragmaTable[i]->ext.url != NULL);
      free(pragmaTable[i]->ext.id);
      free(pragmaTable[i]->ext.url);
      break;
    case typeAccess:
      if(pragmaTable[i]->access.domain)
	free(pragmaTable[i]->access.domain);
      if(pragmaTable[i]->access.path)
	free(pragmaTable[i]->access.path);
      break;
    case typeMeta:
      assert(pragmaTable[i]->meta.key != NULL);
      assert(pragmaTable[i]->meta.value != NULL);
      free(pragmaTable[i]->meta.key);
      free(pragmaTable[i]->meta.value);
      if(pragmaTable[i]->meta.scheme)
	free(pragmaTable[i]->meta.scheme);
      break;

    }

    free(pragmaTable[i]);

  }

  if(pragmaTable)
    free(pragmaTable);
  
}

/**
 * Returns the number of pragmas of the specified type.
 * @param type type of the pragma(s).
 * @return number of pragmas of the specified type.
 */
int WMLScriptPragmaPool::getNumberOfPragmas(const WMLScriptPragmaType type) const {
  int i;
  int c = 0;

  for(i = 0; i < numberOfPragmas; i++)
    if(pragmaTable[i]->type == type) c++;

  return c;
}

/**
 * Retrieves a pragma of the specified type at the specified index.
 * @param type type of the rquired pragma type.
 * @param index of the required pragma.
 * @return pointer at the requested pragma or NULL if the pragma 
 * cannot be found.
 */
WMLScriptPragma *WMLScriptPragmaPool::getPragma(const WMLScriptPragmaType type, const int index) const {
  int i;
  int c = -1;

  for(i = 0; i < numberOfPragmas; i++) {

    if(pragmaTable[i]->type == type) c++;
    if(c == index) return pragmaTable[c];

  }

  return NULL;
}

