/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptValueObject.h,v 1.6 2001/01/27 14:36:42 wouter Exp $
 */

#ifndef __WMLSCRIPTVALUEOBJECT_H__
#define __WMLSCRIPTVALUEOBJECT_H__

/**
 * This enumeration enumerates all the possible types that
 * a value can have in WMLScript. 
 */
typedef enum { 
  /**
   * Boolean type.
   */
  dataBool, 
  /**
   * Integer type.
   */
  dataInt, 
  /**
   * Floating point type.
   */
  dataFloat, 
  /**
   * String type.
   */
  dataString, 
  /**
   * Invalid type (values of this type do not have a value).
   */
  dataInvalid 
} dataType;

/**
 * This struct is a primitive container for values in WMLScript.
 * You generally don't need to use this struct directly; instead,
 * use WMLScriptValueObject's.
 */
typedef struct WMLScriptValueTag {
  /** 
   * The type of the value.
   */
  dataType type;
  /**
   * The actual value of the value.
   */
  union valueTag {
    /**
     * Contains the boolean value if the type is dataBool,
     * where 0 means false.
     */
    char boolValue;
    /**
     * Contains the integer value if the type is dataBool.
     */
    int intValue;
    /**
     * Contains the double precision floating point value if the 
     * type is dataFloat.
     */
    double floatValue;
    /**
     * Contains the string value if the type is dataString.
     */
    char stringValue[1];
  } value;
} WMLScriptValue;

/**
 * This class provides a high-level abstraction for WMLScript
 * values. 
 *
 * WMLScript is a weakly-typed language sporting automatic type
 * conversion. Before using a value, it generally needs to be
 * converted into the proper type. The conversion may not be defined,
 * in which case the resulting value is invalid. Besides acting as a
 * container for a value's type and value, this class also provides 
 * several functions to support automatic conversion between types.
 *
 * This class overrides a number of arithmetic and other operators
 * with implementations that conform to the WMLScript specifications
 * for type conversion and semantics. For example:
 *
 * \code
 * WMLScriptValue v1(3), v2(6), v3("hello");
 * WMLScriptValue v_sum = v1 + v2;  // v_sum.getType() returns dataInteger
 *                                  // v_sum.getIntValue() returns 9
 * WMLScriptValue v_cat = v1 + v3;  // v_cat.getType() returns dataString
 *                                  // v_cat.getStringValue() returns "3hello"
 * \endcode
 *
 * Not all operators have been overridden this way, because not all
 * WMLScript operators can be implemented this way. Most notably, the
 * logical operators || and && have not been implemented, because in
 * WMLScript, the evaluation of the second operand is dependant on the
 * value of the first operand. 
 */
class WMLScriptValueObject {
 public:
  WMLScriptValueObject();
  WMLScriptValueObject(bool v);
  WMLScriptValueObject(int v);
  WMLScriptValueObject(double v);
  WMLScriptValueObject(char *v);
  WMLScriptValueObject(WMLScriptValueObject* orig);
  WMLScriptValueObject(WMLScriptValueObject const &other);
  ~WMLScriptValueObject();
  friend WMLScriptValueObject operator-(const WMLScriptValueObject& in1);
  friend WMLScriptValueObject operator+(const WMLScriptValueObject& in1);
  friend WMLScriptValueObject operator~(const WMLScriptValueObject& in1);
  friend WMLScriptValueObject operator!(const WMLScriptValueObject& in1);
  friend WMLScriptValueObject operator<(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator<=(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator>(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator>=(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator==(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator!=(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator>>(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator<<(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator+ (const WMLScriptValueObject & in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator-(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator*(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator/(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator%(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator|(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator&(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject operator^(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  WMLScriptValueObject const &operator=(const WMLScriptValueObject &other);
  friend WMLScriptValueObject div(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2);
  friend WMLScriptValueObject typeOf(const WMLScriptValueObject& in1);
  friend WMLScriptValueObject isValid(const WMLScriptValueObject& in1);

  dataType getType() const;
  void setValue(bool v);
  void setValue(int v);
  void setValue(double v);
  void setValue(const char *v);
  void setValue();
  int getIntValue() const;
  double getFloatValue() const;
  char *getStringValue() const;
  bool getBoolValue() const;
  void convertToCommon(WMLScriptValueObject *v, dataType minAllowed, dataType maxAllowed);
  void convertToNumber();
  void convertTo(dataType type);
  char *toString();
  bool equals(WMLScriptValueObject *v) const;

 private:
  WMLScriptValue *value;

  void copy(const WMLScriptValueObject &other);
  void destroy();
  void stringToNumber();
  void dprintf(char *s, ...);
  void initialize();
};

#endif
